import math
import pandas as pd
import numpy as np
from math import pow
from scipy import stats, optimize
from six import iteritems
from sys import float_info
from .utils import nanmean, nanstd, nanmin

from .periods import ANNUALIZATION_FACTORS, APPROX_BDAYS_PER_YEAR
from .periods import DAILY, WEEKLY, MONTHLY, QUARTERLY, YEARLY


def _create_unary_vectorized_roll_function(function):

    def unary_vectorized_roll(arr, window, out=None, **kwargs):
        """
        Computes the {human_readable} measure over a rolling window.

        Parameters
        ----------
        arr : array-like
            The array to compute the rolling {human_readable} over.
        window : int
            Size of the rolling window in terms of the periodicity of the data.
        out : array-like, optional
            Array to use as output buffer.
            If not passed, a new array will be created.
        **kwargs
            Forwarded to :func:`~empyrical.{name}`.

        Returns
        -------
        rolling_{name} : array-like
            The rolling {human_readable}.
        """
        allocated_output = out is None

        if len(arr):
            out = function(rolling_window(_flatten(arr), min(len(arr),
                                                             window)).T,
                           out=out,
                           **kwargs)
        else:
            out = np.empty(0, dtype='float64')

        if allocated_output and isinstance(arr, pd.Series):
            out = pd.Series(out, index=arr.index[-len(out):])

        return out

    unary_vectorized_roll.__doc__ = unary_vectorized_roll.__doc__.format(
        name=function.__name__,
        human_readable=function.__name__.replace('_', ' '),
    )

    return unary_vectorized_roll


def _create_binary_vectorized_roll_function(function):

    def binary_vectorized_roll(lhs, rhs, window, out=None, **kwargs):
        """
        Computes the {human_readable} measure over a rolling window.

        Parameters
        ----------
        lhs : array-like
            The first array to pass to the rolling {human_readable}.
        rhs : array-like
            The second array to pass to the rolling {human_readable}.
        window : int
            Size of the rolling window in terms of the periodicity of the data.
        out : array-like, optional
            Array to use as output buffer.
            If not passed, a new array will be created.
        **kwargs
            Forwarded to :func:`~empyrical.{name}`.

        Returns
        -------
        rolling_{name} : array-like
            The rolling {human_readable}.
        """
        allocated_output = out is None

        if window >= 1 and len(lhs) and len(rhs):
            out = function(rolling_window(_flatten(lhs), min(len(lhs),
                                                             window)).T,
                           rolling_window(_flatten(rhs), min(len(rhs),
                                                             window)).T,
                           out=out,
                           **kwargs)
        elif allocated_output:
            out = np.empty(0, dtype='float64')
        else:
            out[()] = np.nan

        if allocated_output:
            if out.ndim == 1 and isinstance(lhs, pd.Series):
                out = pd.Series(out, index=lhs.index[-len(out):])
            elif out.ndim == 2 and isinstance(lhs, pd.Series):
                out = pd.DataFrame(out, index=lhs.index[-len(out):])
        return out

    binary_vectorized_roll.__doc__ = binary_vectorized_roll.__doc__.format(
        name=function.__name__,
        human_readable=function.__name__.replace('_', ' '),
    )

    return binary_vectorized_roll


def annualization_factor(period, annualization):
    """
    Return annualization factor from period entered or if a custom
    value is passed in.

    Parameters
    ----------
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.

    Returns
    -------
    annualization_factor : float
    """
    if annualization is None:
        try:
            factor = ANNUALIZATION_FACTORS[period]
        except KeyError:
            raise ValueError("Period cannot be '{}'. "
                             "Can be '{}'.".format(
                                 period,
                                 "', '".join(ANNUALIZATION_FACTORS.keys())))
    else:
        factor = annualization
    return factor


def _adjust_returns(returns, adjustment_factor):
    """
    Returns the returns series adjusted by adjustment_factor. Optimizes for the
    case of adjustment_factor being 0 by returning returns itself, not a copy!

    Parameters
    ----------
    returns : pd.Series or np.ndarray
    adjustment_factor : pd.Series or np.ndarray or float or int

    Returns
    -------
    adjusted_returns : array-like
    """
    if isinstance(adjustment_factor, (float, int)) and adjustment_factor == 0:
        return returns
    return returns - adjustment_factor


def simple_returns(prices):
    """
    Compute simple returns from a timeseries of prices.

    Parameters
    ----------
    prices : pd.Series, pd.DataFrame or np.ndarray
        Prices of assets in wide-format, with assets as columns,
        and indexed by datetimes.

    Returns
    -------
    returns : array-like
        Returns of assets in wide-format, with assets as columns,
        and index coerced to be tz-aware.
    """
    if isinstance(prices, (pd.DataFrame, pd.Series)):
        out = prices.pct_change().iloc[1:]
    else:
        out = np.diff(prices, axis=0)
        np.divide(out, pirces[:-1], out=out)
    return out


def cum_returns(returns, starting_value=0, out=None):
    """
    Compute cumulative returns from simple returns.

    Parameters
    ----------
    returns : pd.Series, np.ndarray, or pd.DataFrame
        Returns of the strategy as a percentage, noncumulative.
         - Time series with decimal returns.
         - Example::

            2015-07-16   -0.012143
            2015-07-17    0.045350
            2015-07-20    0.030957
            2015-07-21    0.004902

         - Also accepts two dimensional data. In this case, each column is
           cumulated.

    starting_value : float, optional
       The starting returns.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    cumulative_returns : array-like
        Series of cumulative returns.
    """

    if len(returns) < 1:
        return returns.copy()

    nanmask = np.isnan(returns)
    if np.any(nanmask):
        returns = returns.copy()
        returns[nanmask] = 0

    allocated_output = out is None
    if allocated_output:
        out = np.empty_like(returns)

    ## 累计净值计算
    np.add(returns, 1, out=out)
    out.cumprod(axis=0, out=out)

    if starting_value == 0:
        np.subtract(out, 1, out=out)
    else:
        np.multiply(out, starting_value, out=out)

    if allocated_output:
        if returns.ndim == 1 and isinstance(returns, pd.Series):
            out = pd.Series(out, index=returns.index)
        elif isinstance(returns, pd.DataFrame):
            out = pd.DataFrame(
                out,
                index=returns.index,
                columns=returns.columns,
            )

    return out


def cum_returns_final(returns, starting_value=0):
    """
    Compute total returns from simple returns.

    Parameters
    ----------
    returns : pd.DataFrame, pd.Series, or np.ndarray
       Noncumulative simple returns of one or more timeseries.
    starting_value : float, optional
       The starting returns.

    Returns
    -------
    total_returns : pd.Series, np.ndarray, or float
        If input is 1-dimensional (a Series or 1D numpy array), the result is a
        scalar.

        If input is 2-dimensional (a DataFrame or 2D numpy array), the result
        is a 1D array containing cumulative returns for each column of input.
    """

    if len(returns) == 0:
        return np.nan

    if isinstance(returns, pd.DataFrame):
        result = (returns + 1).prod()
    else:
        result = np.nanprod(returns + 1, axis=0)

    if starting_value == 0:
        result -= 1
    else:
        result *= starting_value

    return result


def aggregate_returns(returns, convert_to):
    """
    Aggregates returns by week, month, quarter or year.

    Parameters
    ----------
    returns : pd.Series
       Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    convert_to : str
        Can be 'weekly', 'monthly', or 'yearly'.

    Returns
    -------
    aggregated_returns : pd.Series
    """

    def cumulate_returns(x):
        return cum_returns(x).iloc[-1]

    if convert_to == WEEKLY:
        grouping = [lambda x: x.year, lambda x: x.isocalendar()[1]]
    elif convert_to == MONTHLY:
        grouping = [lambda x: x.year, lambda x: x.month]
    elif convert_to == QUARTERLY:
        grouping = [lambda x: x.year, lambda x: int(math.ceil(x.month / 3.))]
    elif convert_to == YEARLY:
        grouping = [lambda x: x.year]
    else:
        raise ValueError('convert_to must be {}, {} or {}'.format(
            WEEKLY, MONTHLY, YEARLY))

    return returns.groupby(grouping).apply(cumulate_returns)


def max_drawdown(returns, out=None):
    """
    Determines the maximum drawdown of a strategy.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    max_drawdown : float

    Note
    -----
    See https://en.wikipedia.org/wiki/Drawdown_(economics) for more details.
    """
    allocated_output = out is None
    if allocated_output:
        out = np.empty(returns.shape[1:])

    returns_1d = returns.ndim == 1

    if len(returns) < 1:
        out[()] = np.nan
        if returns_1d:
            out = out.item()
        return out

    returns_array = np.asanyarray(returns)

    cumulative = np.empty(
        (returns.shape[0] + 1, ) + returns.shape[1:],
        dtype='float64',
    )
    cumulative[0] = start = 100
    cum_returns(returns_array, starting_value=start, out=cumulative[1:])

    max_return = np.fmax.accumulate(cumulative, axis=0)

    np.nanmin((cumulative - max_return) / max_return, axis=0, out=out)
    if returns_1d:
        out = out.item()
    elif allocated_output and isinstance(returns, pd.DataFrame):
        out = pd.Series(out)

    return out


roll_max_drawdown = _create_unary_vectorized_roll_function(max_drawdown)


def annual_return(returns, period=DAILY, annualization=None):
    """
    Determines the mean annual growth rate of returns. This is equivilent
    to the compound annual growth rate.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Periodic returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.

    Returns
    -------
    annual_return : float
        Annual Return as CAGR (Compounded Annual Growth Rate).

    """

    if len(returns) < 1:
        return np.nan

    ann_factor = annualization_factor(period, annualization)
    num_years = len(returns) / ann_factor
    # Pass array to ensure index -1 looks up successfully.
    ending_value = cum_returns_final(returns, starting_value=1)

    return ending_value**(1 /
                          num_years) - 1  # (1+r)^n = end_value -- r 年化收益  n 年数


def cagr(returns, period=DAILY, annualization=None):
    """
    Compute compound annual growth rate. Alias function for
    :func:`~empyrical.stats.annual_return`

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.
        - See full explanation in :func:`~empyrical.stats.annual_return`.

    Returns
    -------
    cagr : float
        The CAGR value.

    """
    return annual_return(returns, period, annualization)


roll_cagr = _create_unary_vectorized_roll_function(cagr)


def annual_volatility(returns,
                      period=DAILY,
                      alpha=2.0,
                      annualization=None,
                      out=None):
    """
    Determines the annual volatility of a strategy.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Periodic returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    alpha : float, optional
        Scaling relation (Levy stability exponent).
    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    annual_volatility : float
    """
    allocated_output = out is None
    if allocated_output:
        out = np.empty(returns.shape[1:])

    returns_1d = returns.ndim == 1

    if len(returns) < 2:
        out[()] = np.nan
        if returns_1d:
            out = out.item()
        return out

    ann_factor = annualization_factor(period, annualization)
    np.nanstd(returns, ddof=1, axis=0, out=out)
    out = np.multiply(out, ann_factor**(1.0 / alpha), out=out)
    if returns_1d:
        out = out.item()
    return out


roll_annual_volatility = _create_unary_vectorized_roll_function(
    annual_volatility, )


def calmar_ratio(returns, period=DAILY, annualization=None):
    """
    Determines the Calmar ratio, or drawdown ratio, of a strategy.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.


    Returns
    -------
    calmar_ratio : float
        Calmar ratio (drawdown ratio) as float. Returns np.nan if there is no
        calmar ratio.

    Note
    -----
    See https://en.wikipedia.org/wiki/Calmar_ratio for more details.
    """

    max_dd = max_drawdown(returns=returns)
    if max_dd < 0:
        temp = annual_return(returns=returns,
                             period=period,
                             annualization=annualization) / abs(max_dd)
    else:
        return np.nan

    if np.isinf(temp):
        return np.nan

    return temp


def omega_ratio(returns,
                risk_free=0.0,
                required_return=0.0,
                annualization=APPROX_BDAYS_PER_YEAR):
    """Determines the Omega ratio of a strategy.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    risk_free : int, float
        Constant risk-free return throughout the period
    required_return : float, optional
        Minimum acceptance return of the investor. Threshold over which to
        consider positive vs negative returns. It will be converted to a
        value appropriate for the period of the returns. E.g. An annual minimum
        acceptable return of 100 will translate to a minimum acceptable
        return of 0.018.
    annualization : int, optional
        Factor used to convert the required_return into a daily
        value. Enter 1 if no time period conversion is necessary.

    Returns
    -------
    omega_ratio : float

    Note
    -----
    See https://en.wikipedia.org/wiki/Omega_ratio for more details.

    """

    if len(returns) < 2:
        return np.nan

    if annualization == 1:
        return_threshold = required_return
    elif required_return <= -1:
        return np.nan
    else:
        return_threshold = (1 + required_return) ** \
            (1. / annualization) - 1

    returns_less_thresh = returns - risk_free - return_threshold

    numer = sum(returns_less_thresh[returns_less_thresh > 0.0])
    denom = -1.0 * sum(returns_less_thresh[returns_less_thresh < 0.0])

    if denom > 0.0:
        return numer / denom
    else:
        return np.nan


def sharpe_ratio(returns,
                 risk_free=0,
                 period=DAILY,
                 annualization=None,
                 out=None):
    """
    Determines the Sharpe ratio of a strategy.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    risk_free : int, float
        Constant daily risk-free return throughout the period.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    sharpe_ratio : float
        nan if insufficient length of returns or if if adjusted returns are 0.

    Note
    -----
    See https://en.wikipedia.org/wiki/Sharpe_ratio for more details.

    """
    allocated_output = out is None
    if allocated_output:
        out = np.empty(returns.shape[1:])

    return_1d = returns.ndim == 1

    if len(returns) < 2:
        out[()] = np.nan
        if return_1d:
            out = out.item()
        return out

    returns_risk_adj = np.asanyarray(_adjust_returns(returns, risk_free))
    ann_factor = annualization_factor(period, annualization)

    np.multiply(
        np.divide(
            np.nanmean(returns_risk_adj, axis=0),
            np.nanstd(returns_risk_adj, ddof=1, axis=0),
            out=out,
        ),
        np.sqrt(ann_factor),
        out=out,
    )
    if return_1d:
        out = out.item()

    return out


roll_sharpe_ratio = _create_unary_vectorized_roll_function(sharpe_ratio)


def sortino_ratio(returns,
                  required_return=0,
                  period=DAILY,
                  annualization=None,
                  out=None,
                  _downside_risk=None):
    """
    Determines the Sortino ratio of a strategy.

    Parameters
    ----------
    returns : pd.Series or np.ndarray or pd.DataFrame
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    required_return: float / series
        minimum acceptable return
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.
    _downside_risk : float, optional
        The downside risk of the given inputs, if known. Will be calculated if
        not provided.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    sortino_ratio : float or pd.Series

        depends on input type
        series ==> float
        DataFrame ==> pd.Series

    Note
    -----
    See `<https://www.sunrisecapital.com/wp-content/uploads/2014/06/Futures_
    Mag_Sortino_0213.pdf>`__ for more details.

    """
    allocated_output = out is None
    if allocated_output:
        out = np.empty(returns.shape[1:])

    return_1d = returns.ndim == 1

    if len(returns) < 2:
        out[()] = np.nan
        if return_1d:
            out = out.item()
        return out

    adj_returns = np.asanyarray(_adjust_returns(returns, required_return))

    ann_factor = annualization_factor(period, annualization)

    average_annual_return = nanmean(adj_returns, axis=0) * ann_factor
    annualized_downside_risk = (_downside_risk if _downside_risk is not None
                                else downside_risk(returns, required_return,
                                                   period, annualization))
    np.divide(average_annual_return, annualized_downside_risk, out=out)
    if return_1d:
        out = out.item()
    elif isinstance(returns, pd.DataFrame):
        out = pd.Series(out)

    return out


roll_sortino_ratio = _create_unary_vectorized_roll_function(sortino_ratio)


def downside_risk(returns,
                  required_return=0,
                  period=DAILY,
                  annualization=None,
                  out=None):
    """
    Determines the downside deviation below a threshold

    Parameters
    ----------
    returns : pd.Series or np.ndarray or pd.DataFrame
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    required_return: float / series
        minimum acceptable return
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are::

            'monthly':12
            'weekly': 52
            'daily': 252

    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    downside_deviation : float or pd.Series
        depends on input type
        series ==> float
        DataFrame ==> pd.Series

    Note
    -----
    See `<https://www.sunrisecapital.com/wp-content/uploads/2014/06/Futures_
    Mag_Sortino_0213.pdf>`__ for more details, specifically why using the
    standard deviation of the negative returns is not correct.
    """
    allocated_output = out is None
    if allocated_output:
        out = np.empty(returns.shape[1:])

    returns_1d = returns.ndim == 1

    if len(returns) < 1:
        out[()] = np.nan
        if returns_1d:
            out = out.item()
        return out

    ann_factor = annualization_factor(period, annualization)

    downside_diff = np.clip(
        _adjust_returns(
            np.asanyarray(returns),
            np.asanyarray(required_return),
        ),
        np.NINF,
        0,
    )

    np.square(downside_diff, out=downside_diff)
    np.nanmean(downside_diff, axis=0, out=out)
    np.sqrt(out, out=out)
    np.multiply(out, np.sqrt(ann_factor), out=out)

    if returns_1d:
        out = out.item()
    elif isinstance(returns, pd.DataFrame):
        out = pd.Series(out, index=returns.columns)
    return out


roll_downsize_risk = _create_unary_vectorized_roll_function(downside_risk)


def information_ratio(returns, factor_returns):
    """
    Determines the Information ratio of a strategy.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns: float / series
        Benchmark return to compare returns against.

    Returns
    -------
    float
        The information ratio.

    Note
    -----
    See https://en.wikipedia.org/wiki/information_ratio for more details.

    """
    if len(returns) < 2:
        return np.nan

    active_return = _adjust_returns(returns, factor_returns)
    tracking_error = nanstd(active_return, ddof=1)
    if np.isnan(tracking_error):
        return 0.0
    if tracking_error == 0:
        return np.nan
    return nanmean(active_return) / tracking_error


def excess_sharpe(returns, factor_returns, out=None):
    """
    Determines the Excess Sharpe of a strategy.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns: float / series
        Benchmark return to compare returns against.
    out : array-like, optional
        Array to use as output buffer.
        If not passed, a new array will be created.

    Returns
    -------
    excess_sharpe : float

    Note
    -----
    The excess Sharpe is a simplified Information Ratio that uses
    tracking error rather than "active risk" as the denominator.
    """
    allocated_output = out is None
    if allocated_output:
        out = np.empty(returns.shape[1:])

    returns_1d = returns.ndim == 1

    if len(returns) < 2:
        out[()] = np.nan
        if returns_1d:
            out = out.item()
        return out

    active_return = _adjust_returns(returns, factor_returns)
    tracking_error = np.nan_to_num(np.nanstd(active_return, ddof=1, axis=0))

    out = np.divide(
        np.nanmean(active_return, axis=0, out=out),
        tracking_error,
        out=out,
    )
    if returns_1d:
        out = out.item()
    return out


roll_excess_sharpe = _create_binary_vectorized_roll_function(excess_sharpe)


def tail_ratio(returns):
    """Determines the ratio between the right (95%) and left tail (5%).

    For example, a ratio of 0.25 means that losses are four times
    as bad as profits.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
         - See full explanation in :func:`~empyrical.stats.cum_returns`.

    Returns
    -------
    tail_ratio : float
    """

    if len(returns) < 1:
        return np.nan

    returns = np.asanyarray(returns)
    # Be tolerant of nan's
    returns = returns[~np.isnan(returns)]
    if len(returns) < 1:
        return np.nan

    return np.abs(np.percentile(returns, 95)) / \
        np.abs(np.percentile(returns, 5))


def _aligned_series(*many_series):
    """
    Return a new list of series containing the data in the input series, but
    with their indices aligned. NaNs will be filled in for missing values.

    Parameters
    ----------
    many_series : list[pd.Series]

    Returns
    -------
    aligned_series : list[pd.Series]

        A new list of series containing the data in the input series, but
        with their indices aligned. NaNs will be filled in for missing values.

    """
    return [
        series for col, series in iteritems(pd.concat(many_series, axis=1))
    ]


def alpha_beta(returns,
               factor_returns,
               risk_free=0.0,
               period=DAILY,
               annualization=None):
    """Calculates annualized alpha and beta.

    Parameters
    ----------
    returns : pd.Series
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns : pd.Series
         Daily noncumulative returns of the factor to which beta is
         computed. Usually a benchmark such as the market.
         - This is in the same style as returns.
    risk_free : int, float, optional
        Constant risk-free return throughout the period. For example, the
        interest rate on a three month us treasury bill.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are:
            'monthly':12
            'weekly': 52
            'daily': 252
    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.

    Returns
    -------
    float
        Alpha.
    float
        Beta.

    """
    if len(returns) < 2 or len(factor_returns) < 2:
        return np.nan, np.nan

    return alpha_beta_aligned(*_aligned_series(returns, factor_returns),
                              risk_free=risk_free,
                              period=period,
                              annualization=annualization)


def beta_aligned(returns, factor_returns, risk_free=0.0):
    """Calculates beta.

    If they are pd.Series, expects returns and factor_returns have already
    been aligned on their labels.  If np.ndarray, these arguments should have
    the same shape.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns : pd.Series or np.ndarray
         Daily noncumulative returns of the factor to which beta is
         computed. Usually a benchmark such as the market.
         - This is in the same style as returns.
    risk_free : int, float, optional
        Constant risk-free return throughout the period. For example, the
        interest rate on a three month us treasury bill.

    Returns
    -------
    float
        Beta.
    """

    if len(returns) < 2 or len(factor_returns) < 2:
        return np.nan
    # Filter out dates with np.nan as a return value
    joint = np.vstack([_adjust_returns(returns, risk_free), factor_returns])
    joint = joint[:, ~np.isnan(joint).any(axis=0)]
    if joint.shape[1] < 2:
        return np.nan

    cov = np.cov(joint, ddof=0)

    if np.absolute(cov[1, 1]) < 1.0e-30:
        return np.nan

    return cov[0, 1] / cov[1, 1]


def alpha_aligned(returns,
                  factor_returns,
                  risk_free=0.0,
                  period=DAILY,
                  annualization=None,
                  _beta=None):
    """Calculates annualized alpha.

    If they are pd.Series, expects returns and factor_returns have already
    been aligned on their labels.  If np.ndarray, these arguments should have
    the same shape.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns : pd.Series or np.ndarray
         Daily noncumulative returns of the factor to which beta is
         computed. Usually a benchmark such as the market.
         - This is in the same style as returns.
    risk_free : int, float, optional
        Constant risk-free return throughout the period. For example, the
        interest rate on a three month us treasury bill.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are:
            'monthly':12
            'weekly': 52
            'daily': 252
    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.
        - See full explanation in :func:`~empyrical.stats.annual_return`.
    _beta : float, optional
        The beta for the given inputs, if already known. Will be calculated
        internally if not provided.

    Returns
    -------
    float
        Alpha.
    """
    if len(returns) < 2:
        return np.nan

    ann_factor = annualization_factor(period, annualization)

    if _beta is None:
        _beta = beta_aligned(returns, factor_returns, risk_free)

    adj_returns = _adjust_returns(returns, risk_free)
    adj_factor_returns = _adjust_returns(factor_returns, risk_free)
    alpha_series = adj_returns - (_beta * adj_factor_returns)

    return nanmean(alpha_series) * ann_factor


def alpha_beta_aligned(returns,
                       factor_returns,
                       risk_free=0.0,
                       period=DAILY,
                       annualization=None):
    """Calculates annualized alpha and beta.

    If they are pd.Series, expects returns and factor_returns have already
    been aligned on their labels.  If np.ndarray, these arguments should have
    the same shape.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns : pd.Series or np.ndarray
         Daily noncumulative returns of the factor to which beta is
         computed. Usually a benchmark such as the market.
         - This is in the same style as returns.
    risk_free : int, float, optional
        Constant risk-free return throughout the period. For example, the
        interest rate on a three month us treasury bill.
    period : str, optional
        Defines the periodicity of the 'returns' data for purposes of
        annualizing. Value ignored if `annualization` parameter is specified.
        Defaults are:
            'monthly':12
            'weekly': 52
            'daily': 252
    annualization : int, optional
        Used to suppress default values available in `period` to convert
        returns into annual returns. Value should be the annual frequency of
        `returns`.

    Returns
    -------
    float
        Alpha.
    float
        Beta.

    """
    b = beta_aligned(returns, factor_returns, risk_free)
    a = alpha_aligned(returns,
                      factor_returns,
                      risk_free,
                      period,
                      annualization,
                      _beta=b)
    return a, b


# -*- encoding:utf-8 -*-
"""
    择时具体工作者，整合金融时间序列，买入因子，卖出因子，资金类进行
    择时操作，以时间驱动择时事件的发生
"""
import pandas as pd
import numpy as np
import copy
from ultron.ump.factor.buy.base import FactorBuyBase
from ultron.ump.factor.sell.base import FactorSellBase
from ultron.ump.alpha.pick_time_base import PickTimeWorkBase
from ultron.kdutils.progress import MulPidProgress

g_natural_long_task = True


class PickTimeWork(PickTimeWorkBase):
    """择时类"""

    def __init__(self, cap, kl_pd, benchmark, buy_factors, sell_factors):
        """
        :param cap: 资金类AbuCapital实例化对象
        :param kl_pd: 择时时间段交易数据
        :param benchmark: 交易基准对象，AbuBenchmark实例对象
        :param buy_factors: 买入因子序列，序列中的对象为dict，每一个dict针对一个具体因子
        :param sell_factors: 卖出因子序列，序列中的对象为dict，每一个dict针对一个具体因子
        """
        self.capital = cap
        # 回测阶段kl
        self.kl_pd = kl_pd
        # 合并加上回测之前1年的数据，为了生成特征数据
        self.combine_kl_pd = kl_pd  #None  #ABuSymbolPd.combine_pre_kl_pd(self.kl_pd,n_folds=1)
        # 如特别在乎效率性能，打开下面注释的方式，只在g_enable_ml_feature模式下开启, 注释上一行
        #self.combine_kl_pd = ABuSymbolPd.combine_pre_kl_pd(self.kl_pd,
        #                                                    n_folds=1) if ABuEnv.g_enable_ml_feature else None
        # 传递给因子系列，因子内部可有选择性使用
        self.benchmark = benchmark
        # 初始化买入因子列表
        self.init_buy_factors(buy_factors)
        # 初始化卖出因子列表
        self.init_sell_factors(sell_factors)
        # 根据因子是否支持周，月任务属性，筛选周月任务因子对象列表，在初始化时做，提高时间驱动效率
        self.filter_long_task_factors()
        # 择时最终买入卖出行为列表，列表中每一个对象都为Order对象
        self.orders = list()
        # 择时进度条，默认空, 即不打开，不显示择时进度
        self.task_pg = None

    def __str__(self):
        """打印对象显示：买入因子列表＋卖出因子列表"""
        return 'buy_factors:{}\nsell_factors:{}'.format(
            self.buy_factors, self.sell_factors)

    def enable_task_pg(self):
        """启动择时内部任务进度条"""
        if self.kl_pd is not None and hasattr(
                self.kl_pd, 'name') and len(self.kl_pd) > 120:
            self.task_pg = MulPidProgress(
                len(self.kl_pd), 'pick {} times'.format(self.kl_pd.name))
            self.task_pg.init_ui_progress()
            self.task_pg.display_step = 42

    def _week_task(self, today):
        """
        周任务：使用self.week_buy_factors，self.week_sell_factors进行迭代
        不需再使用hasattr进行是否支持判断fit_week
        """
        # 优先执行买入择时因子专属卖出择时因子，而且即使买入因子被锁但附属于买入因子的卖出不能锁
        self._task_attached_sell(today, how='week')

        # 周任务中不建议生成买单，执行卖单，全部在日任务完成，如需判断通过today.exec_week，today.exec_month
        for sell_factor in self.week_sell_factors:
            sell_factor.fit_week(today, self.orders)

        # 执行买入择时因子专属选股因子，决策是否封锁择时买入因子，注意需要从self.buy_factors遍历不是self.week_buy_factors
        self._task_attached_ps(today, is_week=True)

        for buy_factor in self.week_buy_factors:
            if not buy_factor.lock_factor:
                # 如果买入因子没有被封锁执行任务
                buy_factor.fit_week(today)

    def _month_task(self, today):
        """
        月任务：使用self.month_buy_factors，self.month_sell_factors进行迭代
        不需再使用hasattr进行是否支持判断fit_month
        """
        # 优先执行买入择时因子专属卖出择时因子，而且即使买入因子被锁但附属于买入因子的卖出不能锁
        self._task_attached_sell(today, how='month')

        # 月任务中不建议生成买单，执行卖单，全部在日任务完成，如需判断通过today.exec_week，today.exec_month
        for sell_factor in self.month_sell_factors:
            sell_factor.fit_month(today, self.orders)

        # 执行择时因子专属选股因子，决策是否封锁择时买入因子，注意需要从self.buy_factors遍历不是self.month_buy_factors
        self._task_attached_ps(today, is_week=False)

        # 执行带有fit_month的择时买入因子
        for buy_factor in self.month_buy_factors:
            if not buy_factor.lock_factor:
                # 如果买入因子没有被封锁执行任务
                buy_factor.fit_month(today)

    def _day_task(self, today):
        """
        日任务：迭代买入卖出因子序列进行择时
        :param today: 今日的交易数据
        :return:
        """
        # 优先执行买入择时因子专属卖出择时因子，不受买入因子是否被锁的影响
        self._task_attached_sell(today, how='day')

        # 注意回测模式下始终非高频，非当日买卖，不区分美股，A股市场，卖出因子要先于买入因子的执行
        for sell_factor in self.sell_factors:
            # 迭代卖出因子，每个卖出因子针对今日交易数据，已经所以交易单进行择时
            sell_factor.read_fit_day(today, self.orders)

        # 买入因子行为要在卖出因子下面，否则为高频日交易模式
        for buy_factor in self.buy_factors:
            # 如果择时买入因子没有被封锁执行任务
            if not buy_factor.lock_factor:
                # 迭代买入因子，每个因子都对今天进行择时，如果生成order加入self.orders
                order = buy_factor.read_fit_day(today)
                if order and order.order_deal:
                    self.orders.append(order)

    def _task_attached_sell(self, today, how):
        """专属择时买入因子的择时卖出因子任务：日任务择时卖出因子， 周任务择时卖出因子，月任务择时卖出因子"""

        for buy_factor in self.buy_factors:
            # 筛选出当前买入因子所对应的所有单子, 注意这里使用buy_factor_class不是buy_factor，buy_factor带参数做为唯一标示
            factor_orders = list(
                filter(
                    lambda order: order.buy_factor_class == buy_factor.
                    __class__.__name__, self.orders))
            if len(factor_orders) == 0:
                # 当前因子没有对应单子
                continue

            # TODO 不要使用字符串进行eq比对
            for sell_factor in buy_factor.sell_factors:
                if how == 'day':
                    # 所有日任务都要用read_fit_day，且一定存在
                    sell_factor.read_fit_day(today, factor_orders)
                elif how == 'week' and hasattr(sell_factor, 'fit_week'):
                    # 周任务，可选择
                    sell_factor.fit_week(today, factor_orders)
                elif how == 'month' and hasattr(sell_factor, 'fit_month'):
                    # 月任务，可选择
                    sell_factor.fit_month(today, factor_orders)

    def _task_attached_ps(self, today, is_week):
        """专属择时买入因子的选股因子任务：周任务选股因子，月任务选股因子"""
        for buy_factor in self.buy_factors:
            # 不能使用today.exec_week或者today.exec_month来判定，因为有可能都时true
            buy_factor.fit_ps_week(
                today) if is_week else buy_factor.fit_ps_month(today)

    def _task_loop(self, today):
        """
        开始时间驱动，进行日任务，周任务，月任务，
        如果使用自然周，就会在每个周五进行择时操作
        自然月在每个月末最后一天进行择时，否则就以
        天数作为触发条件，这个时候定性任务本身的性质
        只是以时间跨度作为阀值，触发条件
        :param today: 对self.kl_pd apply操作，且axis＝1结果为一天的交易数据
        :return:
        """
        if self.task_pg is not None:
            self.task_pg.show()
        day_cnt = today.key
        # 判断是否执行周任务, 返回结果赋予today对象
        today.exec_week = today.week_task == 1 if g_natural_long_task else day_cnt % 5 == 0
        # 判断是否执行月任务, 返回结果赋予today对象
        today.exec_month = today.month_task == 1 if g_natural_long_task else day_cnt % 20 == 0

        if day_cnt == 0 and not today.exec_week:
            # 如果是择时第一天，且没有执行周任务，需要初始化买入因子专属周任务选股池子
            self._task_attached_ps(today, is_week=True)
        if day_cnt == 0 and not today.exec_month:
            # 如果是择时第一天，且没有执行月任务，需要初始化买入因子专属月任务选股池子
            self._task_attached_ps(today, is_week=False)

        if today.exec_month:
            # 执行因子月任务
            self._month_task(today)
        if today.exec_week:
            # 执行因子周任务
            self._week_task(today)
        # 执行择时因子日任务
        self._day_task(today)

    # noinspection PyTypeChecker
    def fit(self, *args, **kwargs):
        """
            根据交易数据，因子等输入数据，拟合择时
        """
        if g_natural_long_task:
            """如果要进行自然周，自然月择时任务，需要在kl_pd中添加自然周，自然月标记"""
            # 自然周: 每个周五进行标记
            self.kl_pd['week_task'] = np.where(self.kl_pd.date_week == 4, 1, 0)
            """
                自然月: 即前后两个日期，相互减，得到的数 > 60 必然为月末，20140801 - 20140731
                没有使用时间api，因为这样做运行效率快
                self.kl_pd.shift(-1)['date'] - self.kl_pd['date']
                ->
                >>>>
                2014-07-28     1.0
                2014-07-29     1.0
                2014-07-30     1.0
                2014-07-31    70.0
                2014-08-01     3.0
                2014-08-04     1.0
                2014-08-05     1.0
                >>>
                2014-08-22     3.0
                2014-08-25     1.0
                2014-08-26     1.0
                2014-08-27     1.0
                2014-08-28     1.0
                2014-08-29    73.0
                2014-09-02     1.0
                2014-09-03     1.0
                >>>>
            """
            self.kl_pd['month_task'] = np.where(
                self.kl_pd.shift(-1)['date'] - self.kl_pd['date'] > 60, 1, 0)
        # 通过pandas apply进行交易日递进择时
        self.kl_pd.apply(self._task_loop, axis=1)

        if self.task_pg is not None:
            self.task_pg.close_ui_progress()

    def init_sell_factors(self, sell_factors):
        """
        通过sell_factors实例化各个卖出因子
        :param sell_factors: list中元素为dict，每个dict为因子的构造元素，如class，构造参数等
        :return:
        """
        self.sell_factors = list()

        if sell_factors is None:
            return

        for factor_class in sell_factors:
            if factor_class is None:
                continue
            if 'class' not in factor_class:
                # 必须要有需要实例化的类信息
                raise ValueError('factor class key must name class !!!')

            factor_class_cp = copy.deepcopy(factor_class)
            # pop出类信息后剩下的都为类需要的参数
            class_fac = factor_class_cp.pop('class')
            # 整合capital，kl_pd等实例化因子对象
            factor = class_fac(self.capital, self.kl_pd, self.combine_kl_pd,
                               self.benchmark, **factor_class_cp)

            if not isinstance(factor, FactorSellBase):
                # 因子对象类型检测
                raise TypeError('factor must base AbuFactorSellBase')
            # 添加到卖出因子序列
            self.sell_factors.append(factor)

    def init_buy_factors(self, buy_factors):
        """
        通过buy_factors实例化各个买入因子
        :param buy_factors: list中元素为dict，每个dict为因子的构造元素，如class，构造参数等
        :return:
        """
        self.buy_factors = list()

        if buy_factors is None:
            return

        for factor_class in buy_factors:
            if factor_class is None:
                continue

            if 'class' not in factor_class:
                # 必须要有需要实例化的类信息
                raise ValueError('factor class key must name class !!!')

            factor_class_cp = copy.deepcopy(factor_class)
            # pop出类信息后剩下的都为类需要的参数
            class_fac = factor_class_cp.pop('class')
            # 整合capital，kl_pd等实例化因子对象
            factor = class_fac(self.capital, self.kl_pd, self.combine_kl_pd,
                               self.benchmark, **factor_class_cp)

            if not isinstance(factor, FactorBuyBase):
                # 因子对象类型检测
                raise TypeError('factor must base AbuFactorBuyBase')
            # 添加到买入因子序列
            self.buy_factors.append(factor)

    def filter_long_task_factors(self):
        """
        根据每一个因子是否有fit_week筛选周任务因子
        根据每一个因子是否有fit_month筛选月任务因子
        在初始化时完成筛选工作，避免在时间序列中迭代
        不断的进行hasattr判断是否支持
        """
        self.week_buy_factors = list(
            filter(lambda buy_factor: hasattr(buy_factor, 'fit_week'),
                   self.buy_factors))
        self.month_buy_factors = list(
            filter(lambda buy_factor: hasattr(buy_factor, 'fit_month'),
                   self.buy_factors))

        self.week_sell_factors = list(
            filter(lambda sell_factor: hasattr(sell_factor, 'fit_week'),
                   self.sell_factors))
        self.month_sell_factors = list(
            filter(lambda sell_factor: hasattr(sell_factor, 'fit_month'),
                   self.sell_factors))



# -*- coding: utf-8 -*-

from statistics import mode
import unittest

import numpy as np
import pandas as pd

from ultron.optimize.riskmodel import FactorRiskModel
from ultron.optimize.riskmodel import FullRiskModel

class RiskModel(unittest.TestCase):

    def setUp(self):
        self.factor_cov = pd.DataFrame([[0.5, -0.3], [-0.3, 0.7]], columns=['a', 'b'],
                                       index=['a', 'b'])
        self.risk_exp = pd.DataFrame([[0.8, 0.2], [0.5, 0.5], [0.2, 0.8]], columns=['a', 'b'],
                                     index=[1, 2, 3])
        self.idsync = pd.Series([0.1, 0.3, 0.2], index=[1, 2, 3])
        self.sec_cov = self.risk_exp.values @ self.factor_cov.values @ self.risk_exp.values.T \
                       + np.diag(self.idsync.values)
        self.sec_cov = pd.DataFrame(self.sec_cov, columns=[1, 2, 3], index=[1, 2, 3])
    
    def test_full_risk_model(self):
        self.assertEqual(self.sec_cov.shape, (3, 3))
        model = FullRiskModel(self.sec_cov)

        codes = [1, 2]
        res = model.get_cov(codes)
        np.testing.assert_array_almost_equal(res, self.sec_cov.loc[codes, codes].values)

        res = model.get_cov()
        np.testing.assert_array_almost_equal(res, self.sec_cov.values)

    def test_factor_risk_model(self):
        self.assertEqual(self.factor_cov.shape, (2, 2))
        self.assertEqual(self.risk_exp.shape, (3, 2))
        self.assertEqual(self.idsync.shape, (3,))

        model = FactorRiskModel(self.factor_cov,
                                self.risk_exp,
                                self.idsync)
        res = model.get_factor_cov()
        np.testing.assert_array_almost_equal(res, self.factor_cov.values)

        codes = [1, 3]
        res = model.get_risk_exp(codes)
        np.testing.assert_array_almost_equal(res, self.risk_exp.loc[codes, :])
        res = model.get_idsync(codes)
        np.testing.assert_array_almost_equal(res, self.idsync[codes])

        res = model.get_risk_exp()
        np.testing.assert_array_almost_equal(res, self.risk_exp)
        res = model.get_idsync()
        np.testing.assert_array_almost_equal(res, self.idsync)




import os, pdb, itertools, copy, datetime

os.environ['ULTRON_DATA'] = 'keim'

import random
import numpy as np
import pandas as pd
from ultron.env import *
from ultron.factor.genetic.geneticist.operators import custom_transformer
from ultron.factor.genetic.geneticist.engine import Engine

enable_example_env()

market_data = pd.read_csv(os.path.join(g_project_data, 'market_data.csv'),
                          index_col=0)
market_data['trade_date'] = pd.to_datetime(market_data['trade_date'])

tournament_size = 20  ### 初始种群数量
standard_score = 0.75  ### 筛选评估分

custom_params = {
    'tournament_size': tournament_size,
    'standard_score': standard_score,
    'rootid': 'ultron10001'
}


def next_returs_impl(price_data, key, name):
    price_tb = price_data[key].unstack()
    price_tb.fillna(method='pad', inplace=True)
    return_tb = np.log(price_tb.shift(-1) / price_tb)
    return_tb = return_tb.replace([np.inf, -np.inf], np.nan)
    return_tb = return_tb.stack().reindex(price_data.index)
    return_tb.name = name
    return return_tb


next_rets = next_returs_impl(market_data.set_index(['trade_date', 'code']),
                             'closePrice', 'nxt1_ret').reset_index()
next_rets['trade_date'] = pd.to_datetime(next_rets['trade_date'])

sel_factor = pd.read_csv(os.path.join(g_project_data, 'sel_factor.csv'),
                         index_col=0)

total_data = pd.read_csv(os.path.join(g_project_data, 'factor.csv'),
                         index_col=0)
factor_data = total_data[['trade_date', 'code'] +
                         sel_factor['factor'].unique().tolist()]
factor_data['trade_date'] = pd.to_datetime(factor_data['trade_date'])

features = [
    col for col in factor_data.columns
    if col not in ['trade_date', 'code', 'inventory', 'profitratio']
]

total_data = factor_data.merge(next_rets, on=['trade_date', 'code'])


def save_model(gen, rootid, best_programs, custom_params):
    pdb.set_trace()
    print('0--->')


def evaluation(factor_data, total_data, factor_sets, custom_params,
               default_value):
    returns = total_data[['trade_date', 'code', 'nxt1_ret']]
    factor_data = factor_data.reset_index()
    dt = factor_data.merge(returns, on=['trade_date', 'code'])
    factor_ic = dt.groupby(['trade_date']).apply(lambda x: x[
        ['transformed', 'nxt1_ret']].corr(method='spearman').values[0, 1])
    ic = factor_ic.mean()
    return abs(ic)


operators_sets = [
    'MA', 'MADecay', 'MMAX', 'MARGMAX', 'MMIN', 'MARGMIN', 'MRANK',
    'MQUANTILE', 'MSUM', 'MVARIANCE', 'MSTD', 'RSI', 'DELTA', 'SHIFT', 'MCORR'
]
operators_sets = custom_transformer(operators_sets)

gentic = Engine(population_size=10,
                tournament_size=3,
                init_depth=10,
                generations=1000,
                n_jobs=1,
                stopping_criteria=100,
                p_crossover=0.1,
                p_point_mutation=0.5,
                p_subtree_mutation=0.1,
                p_hoist_mutation=0.1,
                p_point_replace=0.1,
                rootid=custom_params['rootid'],
                factor_sets=features,
                standard_score=standard_score,
                operators_set=operators_sets,
                backup_cycle=20,
                convergence=0.002,
                fitness=evaluation,
                save_model=save_model,
                custom_params=custom_params)

gentic.train(total_data=total_data)


# -*- encoding:utf-8 -*-
"""
股价的标准差及其信赖区间，从而确定股价的波动范围及未来走势，利用波带显示股价的安全高低价位，因而也被称为布林带。
其上下限范围不固定，随股价的滚动而变化。布林指标和麦克指标MIKE一样同属路径指标，股价波动在上限和下限的区间之内，
这条带状区的宽窄，随着股价波动幅度的大小而变化，股价涨跌幅度加大时，带状区变宽，涨跌幅度狭小盘整时，带状区则变窄

计算公式
中轨线=N日的移动平均线
上轨线=中轨线+nb_dev * N日的移动标准差
下轨线=中轨线－nb_dev * N日的移动标准差
（nb_dev为参数，可根据股票的特性来做相应的调整，一般默为2）
"""

import numpy as np
import pandas as pd

from ultron.ump.indicator.base import g_calc_type, ECalcType
from ultron.ump.core.helper import pd_rolling_mean, pd_rolling_std


def _calc_boll_from_ta(prices, time_period=20, nb_dev=2):
    """
    使用talib计算boll, 即透传talib.BBANDS计算结果
    :param prices: 收盘价格序列，pd.Series或者np.array
    :param time_period: boll的N值默认值20，int
    :param nb_dev: boll的nb_dev值默认值2，int
    :return: tuple(upper, middle, lower)
    """
    import talib
    if isinstance(prices, pd.Series):
        prices = prices.values

    upper, middle, lower = talib.BBANDS(prices,
                                        timeperiod=time_period,
                                        nbdevup=nb_dev,
                                        nbdevdn=nb_dev,
                                        matype=0)

    return upper, middle, lower


def _calc_boll_from_pd(prices, time_period=20, nb_dev=2):
    """
    通过boll公式手动计算boll
    :param prices: 收盘价格序列，pd.Series或者np.array
    :param time_period: boll的N值默认值20，int
    :param nb_dev: boll的nb_dev值默认值2，int
    :return: tuple(upper, middle, lower)
    """
    # 中轨线 = N日的移动平均线
    middle = pd_rolling_mean(prices,
                             window=time_period,
                             min_periods=time_period)
    # N日的移动标准差
    n_std = pd_rolling_std(prices, window=20, center=False)
    # 上轨线=中轨线+nb_dev * N日的移动标准差
    upper = middle + nb_dev * n_std
    # 下轨线 = 中轨线－nb_dev * N日的移动标准差
    lower = middle - nb_dev * n_std

    # noinspection PyUnresolvedReferences
    return upper.values, middle.values, lower.values


"""通过在ABuNDBase中尝试import talib来统一确定指标计算方式, 外部计算只应该使用calc_boll"""
calc_boll = _calc_boll_from_pd if g_calc_type == ECalcType.E_FROM_PD else _calc_boll_from_ta



# -*- coding: utf-8 -*-

from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySignValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAverageValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityXAverageValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityMACDValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityExpValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityLogValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySqrtValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityPowValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAbsValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAcosValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAcoshValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAsinValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAsinhValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityNormInvValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityCeilValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityFloorValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityRoundValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityDiffValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityRoundValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySigmoidValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityTanhValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySimpleReturnValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityLogReturnValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityMaximumValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityMinimumValueHolder

from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingAverage
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingDecay
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingMax
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingArgMax
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingMin
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingArgMin
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingRank
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingQuantile
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingAllTrue
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingAnyTrue
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingSum
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingVariance
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingStandardDeviation
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingCountedPositive
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingPositiveAverage
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingCountedNegative
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingNegativeAverage
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingPositiveDifferenceAverage
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingNegativeDifferenceAverage
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingRSI
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingLogReturn
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingMovingDrawdown
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingMaxDrawdown
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingSortino
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingSharp
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingCorrelation
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingResidue

__all__ = [
    'SecuritySignValueHolder', 'SecurityAverageValueHolder',
    'SecurityXAverageValueHolder', 'SecurityMACDValueHolder',
    'SecurityExpValueHolder', 'SecurityLogValueHolder',
    'SecuritySqrtValueHolder', 'SecurityPowValueHolder',
    'SecurityAbsValueHolder', 'SecurityAcosValueHolder',
    'SecurityAcoshValueHolder', 'SecurityAsinValueHolder',
    'SecurityAsinhValueHolder', 'SecurityNormInvValueHolder',
    'SecurityCeilValueHolder', 'SecurityFloorValueHolder',
    'SecurityRoundValueHolder', 'SecurityDiffValueHolder',
    'SecurityTanhValueHolder', 'SecuritySigmoidValueHolder',
    'SecuritySimpleReturnValueHolder', 'SecurityLogReturnValueHolder',
    'SecurityMaximumValueHolder', 'SecurityMinimumValueHolder',
    'SecurityMovingAverage', 'SecurityMovingDecay', 'SecurityMovingMax',
    'SecurityMovingArgMax', 'SecurityMovingMin', 'SecurityMovingArgMin',
    'SecurityMovingRank', 'SecurityMovingQuantile', 'SecurityMovingAllTrue',
    'SecurityMovingAnyTrue', 'SecurityMovingSum', 'SecurityMovingVariance',
    'SecurityMovingStandardDeviation', 'SecurityMovingCountedPositive',
    'SecurityMovingPositiveAverage', 'SecurityMovingCountedNegative',
    'SecurityMovingNegativeAverage', 'SecurityMovingPositiveDifferenceAverage',
    'SecurityMovingNegativeDifferenceAverage', 'SecurityMovingRSI',
    'SecurityMovingLogReturn', 'SecurityMovingCorrelation',
    'SecurityMovingMovingDrawdown', 'SecurityMovingMaxDrawdown',
    'SecurityMovingSortino', 'SecurityMovingSharp', 'SecurityMovingResidue'
]



# -*- coding: utf-8 -*-

from typing import Optional
from typing import List
import numpy as np
from . neutralize import neutralize


def factor_processing(raw_factors: np.ndarray,
                      pre_process: Optional[List]=None,
                      risk_factors: Optional[np.ndarray]=None,
                      post_process: Optional[List]=None,
                      groups=None) -> np.ndarray:

    new_factors = raw_factors

    if pre_process:
        for p in pre_process:
            new_factors = p(new_factors, groups=groups)

    if risk_factors is not None:
        risk_factors = risk_factors[:, risk_factors.sum(axis=0) != 0]
        new_factors = neutralize(risk_factors, new_factors, groups=groups)

    if post_process:
        for p in post_process:
            new_factors = p(new_factors, groups=groups)

    return new_factors



# -*- encoding:utf-8 -*-
"""
    卖出择时示例因子：突破卖出择时因子
"""
from ultron.ump.factor.sell.base import FactorSellBase, FactorSellXD, ESupportDirection


class FactorSellBreak(FactorSellBase):
    """示例向下突破卖出择时因子"""

    def _init_self(self, **kwargs):
        """kwargs中必须包含: 突破参数xd 比如20，30，40天...突破"""

        # 向下突破参数 xd， 比如20，30，40天...突破
        self.xd = kwargs['xd']
        # 在输出生成的orders_pd中显示的名字
        self.sell_type_extra = '{}:{}'.format(self.__class__.__name__, self.xd)

    def support_direction(self):
        """支持的方向，只支持正向"""
        return [ESupportDirection.DIRECTION_CAll.value]

    def fit_day(self, today, orders):
        """
        寻找向下突破作为策略卖出驱动event
        :param today: 当前驱动的交易日金融时间序列数据
        :param orders: 买入择时策略中生成的订单序列
        """
        # 今天的收盘价格达到xd天内最低价格则符合条件
        if today.close == self.kl_pd.close[self.today_ind - self.xd +
                                           1:self.today_ind + 1].min():
            for order in orders:
                self.sell_tomorrow(order)


class FactorSellXDBK(FactorSellXD):
    """示例继承AbuFactorBuyXD, 向下突破卖出择时因子"""

    def support_direction(self):
        """支持的方向，只支持正向"""
        return [ESupportDirection.DIRECTION_CAll.value]

    def fit_day(self, today, orders):
        """
        寻找向下突破作为策略卖出驱动event
        :param today: 当前驱动的交易日金融时间序列数据
        :param orders: 买入择时策略中生成的订单序列
        """
        # 今天的收盘价格达到xd天内最低价格则符合条件
        if today.close == self.xd_kl.close.min():
            for order in orders:
                self.sell_tomorrow(order)



# -*- encoding:utf-8 -*-
"""
    交易订单模块
"""

import math

import numpy as np

from ultron.ump.market.futures import FuturesCn
from ultron.ump.market.market import MarketMixin
from ultron.ump.core import env


class OrderMarket(MarketMixin):
    """根据Order对象，设置混入类MarketMixin中symbol_name"""

    def __init__(self, order):
        """
        设置混入类MarketMixin中symbol_name，以获取symbol_market等信息
        :param order: AbuOrder对象
        """
        if isinstance(order, Order):
            self.symbol_name = order.buy_symbol
        else:
            raise TypeError('order is AbuOrder object!!!')


# noinspection PyAttributeOutsideInit
class Order(object):
    """交易订单类"""

    # 多个因子买入条件可能生成几百万个order对象使用__slots__降低内存消耗
    __slots__ = ('order_deal', 'buy_symbol', 'buy_date', 'buy_factor',
                 'buy_factor_class', 'buy_price', 'buy_cnt', 'buy_pos',
                 'sell_date', 'buy_type_str', 'expect_direction', 'sell_type',
                 'keep_days', 'sell_price', 'sell_type_extra', 'ml_features')

    def __init__(self):
        """初始设置只需要将order_deal设置未成交状态"""
        self.order_deal = False

    def fit_buy_order(self, day_ind, factor_object):
        """
        根据买入交易日当当天数据以及买入因子，拟合计算买入订单
        :param day_ind: 买入交易发生的时间索引，即对应self.kl_pd.key
        :param factor_object: ABuFactorBuyBases子类实例对象
        """
        kl_pd = factor_object.kl_pd
        # 要执行买入当天的数据
        kl_pd_buy = kl_pd.iloc[day_ind + 1]
        # 买入因子名称
        factor_name = factor_object.factor_name if hasattr(
            factor_object, 'factor_name') else 'unknown'
        # 日内滑点决策类
        slippage_class = factor_object.slippage_class
        # 仓位管理类设置
        position_class = factor_object.position_class
        # 初始资金，也可修改策略使用剩余资金
        read_cash = factor_object.capital.read_cash
        # 实例化滑点类
        fact = slippage_class(kl_pd_buy, factor_name)
        # 执行fit_price, 计算决策买入价格
        bp = fact.fit()
        # 如果滑点类中决定不买入，撤单子，bp就返回正无穷
        if bp < np.inf:
            """
                实例化仓位管理类
                仓位管理默认保证金比例是1，即没有杠杆，修改ABuPositionBase.g_deposit_rate可提高融资能力，
                如果margin＝2－>ABuPositionBase.g_deposit_rate = 0.5, 即只需要一半的保证金，也可同过构建
                时使用关键字参数完成保证金比例传递
            """
            position = position_class(kl_pd_buy, factor_name,
                                      factor_object.kl_pd.name, bp, read_cash,
                                      **factor_object.position_kwargs)

            market = position.symbol_market  #env.g_market_target if env.g_use_env_market_set else position.symbol_market
            """
                由于模块牵扯复杂，暂时不迁移保证金融资相关模块，期货不使用杠杆，即回测不牵扯资金总量的评估
                if market == EMarketTargetType.E_MARKET_TARGET_FUTURES_CN:
                    deposit_rate = 0.10
                    q_df = AbuFuturesCn().query_symbol(factor_object.kl_pd.name)
                    if q_df is not None:
                        deposit_rate = q_df.min_deposit.values[0]
                    # 重新设置保证金比例
                    position.deposit_rate = deposit_rate
            """
            # 执行fit_position，通过仓位管理计算买入的数量
            bc = position.fit_position(factor_object)
            if np.isnan(bc):
                return

            if market != env.EMarketTargetType.E_MARKET_TARGET_TC:
                # 除了比特币市场外，都向下取整数到最小交易单位个数
                buy_cnt = int(math.floor(bc))
            else:
                # 币类市场可以买非整数个, 保留三位小数
                buy_cnt = round(bc, 3)

            if market == env.EMarketTargetType.E_MARKET_TARGET_US:
                # 美股1
                min_cnt = 1
            elif market == env.EMarketTargetType.E_MARKET_TARGET_TC:
                # 国内一般只支持到0.01个
                min_cnt = 0.01
            elif market == env.EMarketTargetType.E_MARKET_TARGET_CN:
                # A股最小100一手
                min_cnt = 100
                # 向最小的手量看齐
                buy_cnt -= buy_cnt % min_cnt
            elif market == env.EMarketTargetType.E_MARKET_TARGET_FUTURES_CN:
                # 国内期货，查询最少一手单位
                min_cnt = FuturesCn().query_min_unit(factor_object.kl_pd.name)
                # 向最小的手量看齐
                buy_cnt -= buy_cnt % min_cnt
            else:
                raise TypeError(
                    'ABuEnv.g_market_target ERROR, market={}, g_market_target={}'
                    .format(market, env.g_market_target))

            if buy_cnt < min_cnt:
                # 不够买最少单位量
                return

            # 如下生成order内部数据
            self.buy_symbol = kl_pd.name
            # 订单写入买入日期
            self.buy_date = int(kl_pd_buy.date)
            # 订单写入买入因子名字
            self.buy_factor = factor_name
            # 订单对象中添加买入因子类名，和buy_factor不同没有具体参数等唯一key
            self.buy_factor_class = factor_object.__class__.__name__
            # 订单写入买入价格
            self.buy_price = bp
            # 订单写入买入数量
            self.buy_cnt = buy_cnt
            # 订单写入仓位管理类名称
            self.buy_pos = position.__class__.__name__
            # 订单写入买入类型，call or put
            self.buy_type_str = factor_object.buy_type_str
            # 订单写入买入因子期望方向
            self.expect_direction = factor_object.expect_direction

            # 如下卖出信息具体写入在fit_sell_order中
            # 订单卖出时间
            self.sell_date = None
            # 订单卖出类型，keep：持有
            self.sell_type = 'keep'
            # 交易日持有天数
            self.keep_days = 0

            # 订单卖出价格
            self.sell_price = None
            # 订单卖出额外信息
            self.sell_type_extra = ''

            # 订单买入，卖出特征
            self.ml_features = None
            # 订单形成
            self.order_deal = True

    def fit_sell_order(self, day_ind, factor_object):
        """
        根据卖出交易日当当天数据以及卖出因子，拟合计算卖出信息，完成订单
        :param day_ind: 卖出交易发生的时间索引，即对应self.kl_pd.key
        :param factor_object: FactorSellBase子类实例对象
        """

        if self.sell_type != 'keep':
            # 保证外部不需要过滤单子，内部自己过滤已经卖出成交的订单
            return

        # 卖出策略中进行特征合成, 以及ump拦截卖出行为
        if factor_object.make_sell_order(self, day_ind):
            kl_pd_sell = factor_object.kl_pd.iloc[day_ind + 1]

            # 日内滑点决策类
            slippage_class = factor_object.slippage_class
            # 卖出因子名称
            factor_name = factor_object.factor_name if hasattr(
                factor_object, 'factor_name') else 'unknown'
            # 实例化日内滑点决策类，进行具体卖出价格决策
            sell_price = slippage_class(kl_pd_sell, factor_name).fit()

            if sell_price == -np.inf:
                # 如果卖出执行返回负无穷说明无法卖出，例如跌停
                return

            self.sell_price = sell_price
            # 卖出原因其它描述
            sell_type_extra = factor_object.sell_type_extra if hasattr(
                factor_object, 'sell_type_extra') else 'unknown'
            self.sell_type_extra = sell_type_extra
            if self.buy_type_str == 'call':
                # call卖出类型:  win = self.sell_price > self.buy_price
                self.sell_type = 'win' if self.sell_price > self.buy_price else 'loss'
            else:
                # put卖出类型:  loss = self.sell_price > self.buy_price
                self.sell_type = 'loss' if self.sell_price > self.buy_price else 'win'
            # 卖出日期写入单子
            self.sell_date = int(kl_pd_sell.date)

    def __str__(self):
        """打印对象显示：buy_symbol， buy_price， buy_cnt， buy_date，buy_factor，sell_date，sell_type， sell_price"""
        return 'buy Symbol = ' + str(self.buy_symbol) + '\n' \
               + 'buy Prices = ' + str(self.buy_price) + '\n' \
               + 'buy cnt = ' + str(self.buy_cnt) + '\n' \
               + 'buy date = ' + str(self.buy_date) + '\n' \
               + 'buy factor = ' + str(self.buy_factor) + '\n' \
               + 'sell date = ' + str(self.sell_date) + '\n' \
               + 'sell type = ' + str(self.sell_type) + '\n' \
               + 'sell Price = ' + str(self.sell_price) + '\n'

    __repr__ = __str__


# -*- encoding:utf-8 -*-
"""
    日内滑点买入示例实现：均价买入
    最简单的回测买入方式，优点简单，且回测高效，在回测交易
    数量足够多的前提下也能接近实盘
"""
import numpy as np
from ultron.ump.slippage.buy_base import SlippageBuyBase, slippage_limit_up
"""外部修改默认下跌阀值使用如：.slippage.sbm.g_open_down_rate = 0.02"""
g_open_down_rate = 0.07


class SlippageBuyMean(SlippageBuyBase):
    """示例日内滑点均价买入类"""

    @slippage_limit_up
    def fit_price(self):
        """
        取当天交易日的最高最低均价做为决策价格
        :return: 最终决策的当前交易买入价格
        """

        # TODO 基类提取作为装饰器函数，子类根据需要选择是否装饰，并且添加上根据order的call，put明确细节逻辑
        if self.kl_pd_buy.pre_close == 0 or (self.kl_pd_buy.open /
                                             self.kl_pd_buy.pre_close) < (
                                                 1 - g_open_down_rate):
            # 开盘就下跌一定比例阀值，放弃单子
            return np.inf
        # 买入价格为当天均价，即最高，最低的平均，也可使用高开低收平均等方式计算
        self.buy_price = np.mean(
            [self.kl_pd_buy['high'], self.kl_pd_buy['low']])
        # 返回最终的决策价格
        return self.buy_price


# -*- coding: utf-8 -*-

import copy
import numpy as np
cimport numpy as np
cimport cython
from ultron.sentry.Math.Accumulators.IAccumulators cimport Accumulator
from ultron.sentry.Analysis.SeriesValues cimport SeriesValues
from ultron.sentry.Analysis.SecurityValueHolders cimport SecuritySingleValueHolder
from ultron.sentry.Analysis.SecurityValueHolders cimport SecurityBinaryValueHolder
from ultron.sentry.Analysis.SecurityValueHolders cimport build_holder
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingAverage
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingDecay
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingMax
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingArgMax
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingMin
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingArgMin
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingRank
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingQuantile
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingAllTrue
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingAnyTrue
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingSum
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingVariance
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingStandardDeviation
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingCountedPositive
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingPositiveAverage
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingCountedNegative
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingNegativeAverage
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingPositiveDifferenceAverage
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingNegativeDifferenceAverage
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingRSI
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingCorrelation
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingResidue
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingLogReturn
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingSharp
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingSortino
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingMaxDrawdown
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MovingDrawdown
from ultron.sentry.Math.MathConstants cimport NAN



cdef class SecurityMovingAverage(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingAverage, self).__init__(window, MovingAverage, x)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MA}}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)


cdef class SecurityMovingDecay(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingDecay, self).__init__(window, MovingDecay, x)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MADecay}}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)


cdef class SecurityMovingMax(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingMax, self).__init__(window, MovingMax, x)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MMax}}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)


cdef class SecurityMovingArgMax(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingArgMax, self).__init__(window, MovingArgMax, x)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MArgMax}}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)


cdef class SecurityMovingMin(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingMin, self).__init__(window, MovingMin, x)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MMin}}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)


cdef class SecurityMovingArgMin(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingArgMin, self).__init__(window, MovingArgMin, x)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MArgMin}}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)


cdef class SecurityMovingRank(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingRank, self).__init__(window, MovingRank, x)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MRank}}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)


cdef class SecurityMovingQuantile(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingQuantile, self).__init__(window, MovingQuantile, x)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MQuantile}}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)


cdef class SecurityMovingAllTrue(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingAllTrue, self).__init__(window, MovingAllTrue, x)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MAllTrue}}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)


cdef class SecurityMovingAnyTrue(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingAnyTrue, self).__init__(window, MovingAnyTrue, x)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MAnyTrue}}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)


cdef class SecurityMovingSum(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingSum, self).__init__(window, MovingSum, x)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MSum}}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)


cdef class SecurityMovingVariance(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingVariance, self).__init__(window, MovingVariance, x)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MVariance}}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)


cdef class SecurityMovingStandardDeviation(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingStandardDeviation, self).__init__(window, MovingStandardDeviation, x)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MStd}}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)


cdef class SecurityMovingCountedPositive(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingCountedPositive, self).__init__(window, MovingCountedPositive, x)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MNPositive}}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)


cdef class SecurityMovingPositiveAverage(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingPositiveAverage, self).__init__(window, MovingPositiveAverage, x)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MAPositive}}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)


cdef class SecurityMovingCountedNegative(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingCountedNegative, self).__init__(window, MovingCountedNegative, x)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MNNegative}}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)


cdef class SecurityMovingNegativeAverage(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingNegativeAverage, self).__init__(window, MovingNegativeAverage, x)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MANegative}}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)


cdef class SecurityMovingPositiveDifferenceAverage(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingPositiveDifferenceAverage, self).__init__(window, MovingPositiveDifferenceAverage, x)


cdef class SecurityMovingNegativeDifferenceAverage(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingNegativeDifferenceAverage, self).__init__(window, MovingNegativeDifferenceAverage, x)


cdef class SecurityMovingRSI(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingRSI, self).__init__(window, MovingRSI, x)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MRSI}}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)


cdef class SecurityMovingLogReturn(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingLogReturn, self).__init__(window, MovingLogReturn, x)


cdef class SecurityMovingMaxDrawdown(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingMaxDrawdown, self).__init__(window, MovingDrawdown, x)


cdef class SecurityMovingMovingDrawdown(SecuritySingleValueHolder):
    def __init__(self, window, x):
        super(SecurityMovingMovingDrawdown, self).__init__(window, MovingMaxDrawdown, x)


cdef class SecurityMovingSortino(SecurityBinaryValueHolder):
    def __init__(self, window, x, y):
        super(SecurityMovingSortino, self).__init__(window, MovingSortino, x, y)
        
cdef class SecurityMovingSharp(SecurityBinaryValueHolder):
    def __init__(self, window, x, y):
        super(SecurityMovingSharp, self).__init__(window, MovingSharp, x, y)


cdef class SecurityMovingResidue(SecurityBinaryValueHolder):
    def __init__(self, window, x, y):
        super(SecurityMovingResidue, self).__init__(window, MovingResidue, x, y)

    def __str__(self):
        if self._compHolder:
            return "\\mathrm{{MRes}({0}, {1})".format(self._window - self._compHolder.window, str(self._compHolder))
        else:
            return str(self._holderTemplate)

cdef class SecurityMovingCorrelation(SecurityBinaryValueHolder):
    def __init__(self, window, x, y):
        super(SecurityMovingCorrelation, self).__init__(window, MovingCorrelation, x, y)

    def __str__(self):
        return str(self._holderTemplate)




# -*- encoding:utf-8 -*-
"""
    交易执行代理模块
"""

from contextlib import contextmanager
from functools import total_ordering
from enum import Enum
import numpy as np
import pandas as pd
from ultron.ump.trade import trade_execute


class EOrderSameRule(Enum):
    """对order_pd中对order判断为是否相同使用的规则"""
    """order有相同的symbol和买入日期就认为是相同"""
    ORDER_SAME_BD = 0
    """order有相同的symbol, 买入日期，和卖出日期，即不考虑价格，只要日期相同就相同"""
    ORDER_SAME_BSD = 1
    """order有相同的symbol, 买入日期，相同的买入价格，即单子买入时刻都相同"""
    ORDER_SAME_BDP = 2
    """order有相同的symbol, 买入日期, 买入价格, 并且相同的卖出日期和价格才认为是相同，即买入卖出时刻都相同"""
    ORDER_SAME_BSPD = 3


@total_ordering
class OrderPdProxy(object):
    """
        包装交易订单构成的pd.DataFrame对象，外部debug因子的交易结果，寻找交易策略的问题使用，
        支持两个orders_pd的并集，交集，差集，类似set的操作，同时支持相等，不等，大于，小于
        的比较操作，eg如下：

            orders_pd1 = OrderPdProxy(orders_pd1)
            with orders_pd1.proxy_work(orders_pd2) as (order1, order2):
                a = order1 | order2 # 两个交易结果的并集
                b = order1 & order2 # 两个交易结果的交集
                c = order1 - order2 # 两个交易结果的差集(在order1中，但不在order2中)
                d = order2 - order1 # 两个交易结果的差集(在order2中，但不在order1中)
                eq = order1 == order2 # 两个交易结果是否相同
                lg = order1 > order2 # order1唯一的交易数量是否大于order2
                lt = order1 < order2 # order1唯一的交易数量是否小于order2
    """

    def __init__(self, orders_pd, same_rule=EOrderSameRule.ORDER_SAME_BSPD):
        """
        初始化函数需要pd.DataFrame对象，暂时未做类型检测
        :param orders_pd: 回测结果生成的交易订单构成的pd.DataFrame对象
        :param same_rule: order判断为是否相同使用的规则, 默认EOrderSameRule.ORDER_SAME_BSPD
                          即：order有相同的symbol和买入日期和相同的卖出日期和价格才认为是相同
        """
        # 需要copy因为会添加orders_pd的列属性等
        self.orders_pd = orders_pd.copy()
        self.same_rule = same_rule
        # 并集, 交集, 差集运算结果存储
        self.op_result = None
        self.last_op_metrics = {}

    @contextmanager
    def proxy_work(self, orders_pd):
        """
        传人需要比较的orders_pd，构造OrderPdProxy对象，返回使用者，
        对op_result进行统一分析
        :param orders_pd: 回测结果生成的交易订单构成的pd.DataFrame对象
        :return:
        """
        # 运算集结果重置
        self.op_result = None
        # 实例化比较的OrderPdProxy对象
        other = OrderPdProxy(orders_pd)
        try:
            yield self, other
        finally:
            if isinstance(self.op_result, pd.DataFrame):
                # 如果有并集, 交集, 差集运算结果存储，
                from ultron.ump.metrics.metrics_base import MetricsBase
                metrics = MetricsBase(self.op_result, None, None, None)
                metrics.fit_metrics_order()

                self.last_op_metrics['win_rate'] = metrics.win_rate
                self.last_op_metrics['gains_mean'] = metrics.gains_mean
                self.last_op_metrics['losses_mean'] = metrics.losses_mean
                self.last_op_metrics['sum_profit'] = self.op_result[
                    'profit'].sum()
                self.last_op_metrics['sum_profit_cg'] = self.op_result[
                    'profit_cg'].sum()

    def __and__(self, other):
        """ & 操作符的重载，计算两个交易集的交集"""
        # self.op = 'intersection(order1 & order2)'
        self.op_result = intersection_in_2orders(self.orders_pd,
                                                 other.orders_pd,
                                                 same_rule=self.same_rule)
        return self.op_result

    def __or__(self, other):
        """ | 操作符的重载，计算两个交易集的并集"""
        # self.op = 'union(order1 | order2)'
        self.op_result = union_in_2orders(self.orders_pd, other.orders_pd)
        return self.op_result

    def __sub__(self, other):
        """ - 操作符的重载，计算两个交易集的差集"""
        self.op_result = difference_in_2orders(self.orders_pd,
                                               other.orders_pd,
                                               same_rule=self.same_rule)
        return self.op_result

    def __eq__(self, other):
        """ == 操作符的重载，计算两个交易集的是否相同"""
        return (self - other).empty and (other - self).empty

    def __gt__(self, other):
        """ > 操作符的重载，计算两个交易集的大小, 类被total_ordering装饰，可以支持lt等操作符"""
        unique_cnt = find_unique_group_symbol(self.orders_pd).shape[0]
        other_unique_cnt = find_unique_group_symbol(other.orders_pd).shape[0]
        return unique_cnt > other_unique_cnt


def union_in_2orders(orders_pd, other_orders_pd):
    """
    并集：分析因子或者参数问题时使用，debug策略问题时筛选出两个orders_pd中所有不同的交易，
    注意这里不认为在相同的交易日买入相同的股票，两笔交易就一样，这里只是两个orders_pd合并
    后使用drop_duplicates做了去除完全一样的order，即结果为并集：
    orders_pd | cmp_orders_pd或orders_pd.union(cmp_orders_pd)
    :param orders_pd: 回测结果生成的交易订单构成的pd.DataFrame对象
    :param other_orders_pd: 回测结果生成的交易订单构成的pd.DataFrame对象
    :return: orders_pd | cmp_orders_pd
    """
    orders_pd = orders_pd.append(other_orders_pd)
    orders_pd = orders_pd.drop_duplicates()
    return orders_pd


def _same_pd(order, other_orders_pd, same_rule):
    """
    根据same_rule的规则从orders_pd和other_orders_pd中返回相同的df

    :param order: orders_pd中的一行order记录数据
    :param other_orders_pd: 回测结果生成的交易订单构成的pd.DataFrame对象
    :param same_rule: order判断为是否相同使用的规则
    :return: 从orders_pd和other_orders_pd中返回相同的df
    """
    symbol = order.symbol
    buy_day = order['buy_date']
    buy_price = order['buy_price']

    sell_day = order['sell_date']
    sell_price = order['sell_price']

    if same_rule == EOrderSameRule.ORDER_SAME_BD:
        # 只根据买入时间和买入symbol确定是否相同，即认为在相同的交易日买入相同的股票，两笔交易就一样，忽略其它所有order中的因素
        same_pd = other_orders_pd[(other_orders_pd['symbol'] == symbol)
                                  & (other_orders_pd['buy_date'] == buy_day)]
    elif same_rule == EOrderSameRule.ORDER_SAME_BSD:
        # 根据买入时间，卖出时间和买入symbol确定是否相同
        same_pd = other_orders_pd[(other_orders_pd['symbol'] == symbol)
                                  & (other_orders_pd['buy_date'] == buy_day)
                                  & (other_orders_pd['sell_date'] == sell_day)]
    elif same_rule == EOrderSameRule.ORDER_SAME_BDP:
        # 根据买入时间，买入价格和买入symbol确定是否相同
        same_pd = other_orders_pd[(other_orders_pd['symbol'] == symbol)
                                  & (other_orders_pd['buy_date'] == buy_day)
                                  &
                                  (other_orders_pd['buy_price'] == buy_price)]
    elif same_rule == EOrderSameRule.ORDER_SAME_BSPD:
        # 根据买入时间，卖出时间, 买入价格和卖出价格和买入symbol确定是否相同
        same_pd = other_orders_pd[(other_orders_pd['symbol'] == symbol)
                                  & (other_orders_pd['buy_date'] == buy_day)
                                  & (other_orders_pd['sell_date'] == sell_day)
                                  & (other_orders_pd['buy_price'] == buy_price)
                                  & (other_orders_pd['sell_price']
                                     == sell_price)]
    else:
        raise TypeError('same_rule type is {}!!'.format(same_rule))
    return same_pd


def intersection_in_2orders(orders_pd,
                            other_orders_pd,
                            same_rule=EOrderSameRule.ORDER_SAME_BSPD):
    """
    交集: 分析因子或者参数问题时使用，debug策略问题时筛选出两个orders_pd中相同的交易，
    即结果为交集：orders_pd & cmp_orders_pd或orders_pd.intersection(cmp_orders_pd)
    :param orders_pd: 回测结果生成的交易订单构成的pd.DataFrame对象
    :param other_orders_pd: 回测结果生成的交易订单构成的pd.DataFrame对象
    :param same_rule: order判断为是否相同使用的规则, 默认EOrderSameRule.ORDER_SAME_BSPD
                          即：order有相同的symbol和买入日期和相同的卖出日期和价格才认为是相同
    :return: orders_pd & cmp_orders_pd
    """

    def _intersection(order):
        same_pd = _same_pd(order, other_orders_pd, same_rule)
        if same_pd.empty:
            # 如果是空，说明不相交
            return False
        # 相交, intersection=1，是交集
        return True

    orders_pd['intersection'] = orders_pd.apply(_intersection, axis=1)
    return orders_pd[orders_pd['intersection'] == 1]


def difference_in_2orders(orders_pd,
                          other_orders_pd,
                          same_rule=EOrderSameRule.ORDER_SAME_BSPD):
    """
    差集: 分析因子或者参数问题时使用，debug策略问题时筛选出两个orders_pd的不同交易，
    注意返回的结果是存在orders_pd中的交易，但不在cmp_orders_pd中的交易，即结果
    为差集：orders_pd - cmp_orders_pd或orders_pd.difference(cmp_orders_pd)
    :param orders_pd: 回测结果生成的交易订单构成的pd.DataFrame对象
    :param other_orders_pd: 回测结果生成的交易订单构成的pd.DataFrame对象
    :param same_rule: order判断为是否相同使用的规则, 默认EOrderSameRule.ORDER_SAME_BSPD
                      即：order有相同的symbol和买入日期和相同的卖出日期和价格才认为是相同
    :return: orders_pd - cmp_orders_pd
    """

    def _difference(order):
        same_pd = _same_pd(order, other_orders_pd, same_rule)
        if same_pd.empty:
            # 没有相同的说明是差集
            return True
        # 有相同的说明不是差集
        return False

    orders_pd['difference'] = orders_pd.apply(_difference, axis=1)
    return orders_pd[orders_pd['difference'] == 1]


def find_unique_group_symbol(order_pd):
    """
    按照'buy_date', 'symbol'分组后，只筛选组里的第一个same_group.iloc[0]
    :param order_pd:
    :return:
    """

    def _find_unique_group_symbol(same_group):
        # 只筛选组里的第一个, 即同一个交易日，对一个股票的交易只保留一个order
        return same_group.iloc[0]

    # 按照'buy_date', 'symbol'分组后apply same_handle
    order_pds = order_pd.groupby(['buy_date',
                                  'symbol']).apply(_find_unique_group_symbol)
    return order_pds


def find_unique_symbol(order_pd, same_rule=EOrderSameRule.ORDER_SAME_BSPD):
    """
    order_pd中如果一个buy_date对应的一个symbol有多条交易记录，过滤掉，
    注意如果在对应多条记录中保留一个，使用find_unique_group_symbol
    :param order_pd: 回测结果生成的交易订单构成的pd.DataFrame对象
    :param same_rule: order判断为是否相同使用的规则, 默认EOrderSameRule.ORDER_SAME_BSPD
                  即：order有相同的symbol和买入日期和相同的卖出日期和价格才认为是相同
    """

    def _find_unique_symbol(order):
        """根据order的symbol和buy_date在原始order_pd中进行复合条件筛选，结果same_pd如果只有1个就唯一，否则就是重复的"""
        same_pd = _same_pd(order, order_pd, same_rule)
        if same_pd.empty or same_pd.shape[0] == 1:
            return False
        # 同一天一个symbol有多条记录的一个也没留，都过滤
        return True

    same_mark = order_pd.apply(_find_unique_symbol, axis=1)
    return order_pd[same_mark == 0]


def trade_summary(orders, kl_pd, show_info=True):
    """
    主要将AbuOrder对象序列转换为pd.DataFrame对象orders_pd，以及将
    交易单子时间序列转换交易行为顺序序列，绘制每笔交易的细节交易图，以及
    简单文字度量输出
    :param orders: AbuOrder对象序列
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :param draw: 是否可视化交易细节图示
    :param show_info: 是否输出交易文字信息
    """

    # Order对象序列转换为pd.DataFrame对象orders_pd
    orders_pd = trade_execute.make_orders_pd(orders, kl_pd)
    # 交易单子时间序列转换交易行为顺序序列
    action_pd = trade_execute.transform_action(orders_pd)

    summary = ''

    if show_info:
        # simple的意思是没有计算交易费用
        simple_profit = 'simple profit: {} \n'.format(
            trade_execute.calc_simple_profit(orders, kl_pd))
        summary += simple_profit

        # 每笔交易收益期望
        mean_win_profit = 'mean win profit {} \n'.format(
            np.mean(orders_pd[orders_pd.result == 1]['profit']))
        summary += mean_win_profit

        # 每笔交易亏损期望
        mean_loss_profit = 'mean loss profit {} \n'.format(
            np.mean(orders_pd[orders_pd.result == -1]['profit']))
        summary += mean_loss_profit

        # 盈利笔数
        win_cnt = 0 if len(orders_pd[orders_pd.result == 1].result.value_counts().values) <= 0 else \
            orders_pd[orders_pd.result == 1].result.value_counts().values[0]

        # 亏损笔数
        loss_cnt = 0 if len(orders_pd[orders_pd.result == -1].result.value_counts().values) <= 0 else \
            orders_pd[orders_pd.result == -1].result.value_counts().values[0]

        # 胜率
        win_rate = 'win rate ' + str('*@#')
        if win_cnt + loss_cnt > 0:
            win_rate = 'win rate: {}%'.format(
                float(win_cnt) / float(float(loss_cnt) + float(win_cnt)))
        summary += win_rate

    return orders_pd, action_pd, summary


# -*- encoding:utf-8 -*-
"""
    字符工具模块
"""


def digit_str(item):
    """
        从第一个字符开始删除，直到所有字符都是数字为止，或者item长度 < 2
        eg:
            input:  ABuStrUtil.digit_str('sh000001')
            output: 000001

            input:  ABuStrUtil.digit_str('shszsh000001')
            output: 000001
    :param item: 字符串对象
    :return: 过滤head字母的字符串对象
    """
    while True:
        if item.isdigit():
            break
        if len(item) < 2:
            break
        item = item[1:]
    return item


# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
from ultron.strategy.builder.factor import *
from ultron.strategy.builder.universe import Universe


def calc_weighted_signal(signals, factor_list, factor_weight):
    index = signals[list(signals.keys())[0]].index
    columns = signals[list(signals.keys())[0]].columns

    weights = factor_weight

    rval = pd.DataFrame(0, index=index, columns=columns)
    for i, strategy in enumerate(factor_list):
        rval += weights[i] * signals[strategy]
    return rval


def calc_1k_signal(signals, factor_list, factor_weight, oneover_k_equal=True):
    if not oneover_k_equal:
        weights = factor_weight
    else:
        weights = len(factor_list) * [1]

    index = signals[list(signals.keys())[0]].index
    columns = signals[list(signals.keys())[0]].columns

    rval = pd.DataFrame(0, index=index, columns=columns)
    for i, strategy in enumerate(factor_list):
        score = signals[strategy]
        score.fillna(0, inplace=True)
        score = np.sign(score)
        rval += weights[i] * score

    # nominal invest capital scale: 50% long, 50% short
    pos = rval.copy()
    pos[pos < 0] = np.nan
    pos.fillna(0, inplace=True)
    pos = pos.div(np.sum(pos, axis=1) + 1e-16, axis=0) * .5
    neg = rval.copy()
    neg[neg > 0] = np.nan
    neg.fillna(0, inplace=True)
    neg = neg.div(np.sum(neg, axis=1) + 1e-16, axis=0) * .5

    rval = pos + neg
    return rval


def calc_composite_signal(signals,
                          factor_list,
                          factor_weight,
                          composite_equal=True):
    if not composite_equal:
        weights = factor_weight
    else:
        weights = len(factor_weight) * [1]

    index = signals[list(signals.keys())[0]].index
    columns = signals[list(signals.keys())[0]].columns

    rval = pd.DataFrame(0, index=index, columns=columns)
    for i, strategy in enumerate(factor_list):
        score = signals[strategy]
        score.fillna(0, inplace=True)
        score = np.sign(score)
        rval += weights[i] * score

    # calc target position weights
    rval = rval.div(rval.abs().sum(axis=1) + 1e-16, axis=0)
    return rval


def calc_factors_score_signal(signals,
                              factor_list,
                              factor_weight,
                              investment_ratio=0.4):

    def get_nth_largest_value(df, n):
        argout = df.apply(lambda x: x.nlargest(n).values[-1], axis=1)
        return argout

    def get_nth_smallest_value(df, n):
        argout = df.apply(lambda x: x.nsmallest(n).values[-1], axis=1)
        return argout

    weights = factor_weight

    index = signals[list(signals.keys())[0]].index
    columns = signals[list(signals.keys())[0]].columns

    rval = pd.DataFrame(0, index=index, columns=columns)
    for i, strategy in enumerate(factor_list):
        score = cross_section_rank(signals[strategy])
        rval += weights[i] * score

    nrow, ncol = rval.shape
    count = rval.count(axis=1)
    num = count * investment_ratio  #self.signalMultiFactorsScoreInvestmentRatio
    num = num.round(0)
    num = int(num.tolist()[-1])

    i0 = count[count >= num * 2].index.tolist()[0]

    # upper part
    bnd = pd.Series(np.nan, index=count.index)
    bnd[i0:] = get_nth_largest_value(rval[i0:], num)
    bnd = bnd.repeat(ncol).values.reshape((nrow, ncol))
    pos = rval.copy()
    pos[pos < bnd] = np.nan
    pos.fillna(0, inplace=True)

    # lower part
    bnd = pd.Series(np.nan, index=count.index)
    bnd[i0:] = get_nth_smallest_value(rval[i0:], num)
    bnd = bnd.repeat(ncol).values.reshape((nrow, ncol))
    neg = rval.copy()
    neg[neg > bnd] = np.nan
    neg.fillna(0, inplace=True)

    rval = np.sign(pos) - np.sign(neg)
    rval = rval.div(rval.abs().sum(axis=1) + 1e-16, axis=0)
    return rval


def build_qs_signal(data,
                    split_strategy_type,
                    bound,
                    winsize,
                    rank,
                    upper,
                    lower,
                    zscore_normalization=True,
                    equal_weighted=True,
                    nominal_invest_weight=True,
                    qs_row_scale=True):
    universe = Universe.all(data)
    if zscore_normalization:
        data = data.copy()
        data = ts_zscore_normalization(data=data, bound=bound, winsize=winsize)
    upp = qs_percentile(data, universe, (upper - 1) / rank, upper / rank)
    upp.fillna(0, inplace=True)
    upp = abs(upp)

    lop = qs_percentile(data, universe, (lower - 1) / rank, lower / rank)
    lop.fillna(0, inplace=True)
    lop = abs(lop)

    if equal_weighted:
        upp = np.sign(upp)
        lop = np.sign(lop)
    rval = upp - lop

    if nominal_invest_weight:
        weights = cross_section_universe_equal(universe)
    else:
        weights = pd.DataFrame(1, index=data.index, columns=data.columns)
    rval = rval * weights

    if qs_row_scale:
        rval = cross_section_scale(rval)
    rval[~universe] = np.nan
    return rval


def build_cs_signal(data,
                    bound,
                    winsize,
                    zscore_normalization=True,
                    equal_weighted=True):
    universe = Universe.all(data)
    if zscore_normalization:
        data = ts_zscore_normalization(data=data, bound=bound, winsize=winsize)
    rval = cross_section_demean_rank(data, universe=universe)
    if equal_weighted:
        rval = np.sign(rval)
    rval = cross_section_nan_scale(rval)
    return rval


def build_ts_signal(data,
                    bound,
                    winsize,
                    zscore_normalization=True,
                    nominal_invest_weight=True,
                    ts_row_scale=True):
    universe = Universe.all(data)
    if zscore_normalization:
        x = ts_zscore_normalization(data=data, bound=bound, winsize=winsize)
    else:
        x = data.copy()

    if nominal_invest_weight:
        weights = cross_section_universe_equal(universe=universe)
    else:
        weights = pd.DataFrame(data=np.nan,
                               index=universe.index,
                               columns=universe.columns)
        ones = pd.DataFrame(data=1.0,
                            index=universe.index,
                            columns=universe.columns)
        weights[universe] = ones[universe]

    rval = x * weights
    if ts_row_scale:
        rval = cross_section_scale(rval)
    return rval


# -*- encoding:utf-8 -*-
"""
    多任务子进程拷贝跟随主进程设置模块
"""

import functools
import imp
from ultron.ump.core.fixes import filter, signature, Parameter


def add_process_env_sig(func):
    """
    初始化装饰器时给被装饰函数添加env关键字参数，在wrapper中将env对象进行子进程copy
    由于要改方法签名，多个装饰器的情况要放在最下面
    :param func:
    :return:
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # env = kwargs.pop('env', None)
        if 'env' in kwargs:
            """
                实际上linux, mac os上并不需要进行进程间模块内存拷贝，
                子进程fork后携带了父进程的内存信息，win上是需要的，
                暂时不做区分，都进行进程间的内存拷贝，如特别在乎效率的
                情况下基于linux系统，mac os可以不需要拷贝，如下：
                if kwargs['env'] is not None and not ABuEnv.g_is_mac_os:
                    # 只有windows进行内存设置拷贝
                    env.copy_process_env()
            """
            # if kwargs['env'] is not None and not ABuEnv.g_is_mac_os:
            env = kwargs.pop('env', None)
            if env is not None:
                # 将主进程中的env拷贝到子进程中
                env.copy_process_env()
        return func(*args, **kwargs)

    # 获取原始函数参数签名，给并行方法添加env参数
    sig = signature(func)

    if 'env' not in list(sig.parameters.keys()):
        parameters = list(sig.parameters.values())
        # 通过强制关键字参数，给方法加上env
        parameters.append(
            Parameter('env', Parameter.KEYWORD_ONLY, default=None))
        # wrapper的__signature__进行替换
        wrapper.__signature__ = sig.replace(parameters=parameters)

    return wrapper


class EnvProcess(object):
    """多任务主进程内存设置拷贝执行者类"""

    def __init__(self):
        """迭代注册了的需要拷贝内存设置的模块，通过筛选模块中以g_或者_g_开头的属性，将这些属性拷贝为类属性变量"""
        for module in self.register_module():
            # 迭代注册了的需要拷贝内存设置的模块, 筛选模块中以g_或者_g_开头的, 且不能callable，即不是方法
            sig_env = list(
                filter(
                    lambda _sig: not callable(_sig) and
                    (_sig.startswith('g_') or _sig.startswith('_g_')),
                    dir(module)))

            module_name = module.__name__
            # map(lambda sig: setattr(self, '{}_{}'.format(module_name, sig), module.__dict__[sig]), sig_env)
            for sig in sig_env:
                # 模块中的属性拷贝为类属性变量，key＝module_name_sig
                setattr(self, '{}_{}'.format(module_name, sig),
                        module.__dict__[sig])

    # noinspection PyMethodMayBeStatic
    def register_module(self):
        """
        注册需要拷贝内存的模块，不要全局模块注册，否则很多交叉引用，也不要做为类变量存储否则多进程传递pickle时会出错
        :return:
        """
        from ultron.ump.beta import atr_position, position_base
        from ultron.ump.core import env
        from ultron.ump.similar import corrcoef
        from ultron.ump.slippage import buy_base, buy_mean, sell_base
        from ultron.ump.trade import ml_feature
        from ultron.ump.ump import manager, main_base, edge_base
        #from ultron.ump.technical import
        from ultron.ump.market import market
        from ultron.ump.alpha import pick_time_worker
        from ultron.ump.factor.sell import close_atrn_stop, pre_atrn_stop
        from ultron.kdutils.progress import Progress
        #from ..PickStockBu import ABuPickSimilarNTop
        #from ..UtilBu import ABuProgress

        # TODO 将每个模块中全局设置放在一个模块配置代码文件中，这里只将所有模块配置代码文件加载
        return [
            atr_position, position_base, env, corrcoef, Progress, buy_base,
            sell_base, buy_mean, main_base, edge_base, ml_feature, manager,
            pick_time_worker, close_atrn_stop, market, pre_atrn_stop
        ]

    def copy_process_env(self):
        """为子进程拷贝主进程中的设置执行，在add_process_env_sig装饰器中调用，外部不应主动使用"""
        for module in self.register_module():
            # 迭代注册了的需要拷贝内存设置的模块, 筛选模块中以g_或者_g_开头的, 且不能callable，即不是方法
            sig_env = list(
                filter(
                    lambda sig: not callable(sig) and
                    (sig.startswith('g_') or sig.startswith('_g_')),
                    dir(module)))
            module_name = module.__name__
            for _sig in sig_env:
                # 格式化类变量中对应模块属性的key
                name = '{}_{}'.format(module_name, _sig)
                # 根据应模块属性的key（name）getattr获取属性值
                val = getattr(self, name)
                # 为子模块内存变量进行值拷贝
                module.__dict__[_sig] = val
                # print(name, val)

    def __str__(self):
        """打印对象显示：注册需要拷贝内存的模块中在AbuEnvProcess对象属性的映射key值，以及value值"""

        str_dict = dict()
        for module in self.register_module():
            sig_env = list(
                filter(
                    lambda sig: not callable(sig) and
                    (sig.startswith('g_') or sig.startswith('_g_')),
                    dir(module)))
            module_name = module.__name__
            for _sig in sig_env:
                # format对象属性的映射key值
                name = '{}_{}'.format(module_name, _sig)
                # 根据映射key值getattr出value值
                attr_str = getattr(self, name)
                str_dict[name] = attr_str
        return str(str_dict)

    __repr__ = __str__



# -*- coding: utf-8 -*-
import time
import numpy as np
from ultron.sentry.Utilities.Asserts import pyFinAssert
from ultron.sentry.Utilities.Asserts import pyFinWarning
from ultron.sentry.Utilities.Asserts import isClose

__all__ = ['pyFinAssert',
           'pyFinWarning',
           'isClose']


def print_timing(func):
    def wrapper(*arg):
        t1 = time.time()
        res = func(*arg)
        t2 = time.time()
        return t2 - t1, res
    return wrapper









from utilities.test_runner import Runner


if __name__ == '__main__':
    from ultron.tests.factor.data.neutralize import Neutralize
    from ultron.tests.factor.data.quantile import Quantile
    from ultron.tests.factor.data.standardize import Standardize
    from ultron.tests.factor.data.winsorize import Winsorize
    from ultron.tests.optimize.linearbuild import LinearBuild
    from ultron.tests.optimize.longshortbuild import LongShortBuild
    from ultron.tests.optimize.meanvariancebuild import MeanVarianceBuild
    from ultron.tests.optimize.optimizers import Optimizers
    from ultron.tests.optimize.risk_model import RiskModel
    from ultron.tests.optimize.model.loader import Loader
    from ultron.tests.optimize.model.modelbase import ModelBase
    from ultron.tests.optimize.model.linearmodel import LinearModel
    from ultron.tests.optimize.model.treemodel import TreeModel

    runner = Runner([Neutralize,Quantile,Standardize,Winsorize,LinearBuild,
                    LongShortBuild,MeanVarianceBuild,Optimizers,RiskModel,
                    Loader,ModelBase,LinearModel,TreeModel])
    runner.run()


# -*- encoding:utf-8 -*-
"""
    择时与选股抽象基类
"""
from abc import ABCMeta, abstractmethod
from ultron.ump.core.fixes import six
from ultron.ump.core.base import ParamBase


class PickTimeWorkBase(six.with_metaclass(ABCMeta, ParamBase)):
    """择时抽象基类"""

    @abstractmethod
    def fit(self, *args, **kwargs):
        """
        fit在整个项目中的意义为开始对象最重要的工作，
        对于择时对象即为开始择时操作，或者从字面理解
        开始针对交易数据进行拟合择时操作
        """
        pass

    @abstractmethod
    def init_sell_factors(self, *args, **kwargs):
        """
        初始化择时卖出因子
        """
        pass

    @abstractmethod
    def init_buy_factors(self, *args, **kwargs):
        """
        初始化择时买入因子
        """
        pass


# -*- encoding:utf-8 -*-
"""
    日内滑点买入价格决策基础模块：暂时迁移简单实现方式，符合回测需求，如迁移实盘模块
    需添加日内择时策略，通过日内分钟k线，实现日内分钟k线择时，更微观的
    实现日内择时滑点功能，不考虑大资金的冲击成本及系统外的大幅滑点
"""

from abc import ABCMeta, abstractmethod
import functools

import numpy as np

from ultron.ump.core.fixes import six


class SlippageBuyBase(six.with_metaclass(ABCMeta, object)):
    """非高频日内滑点买入决策抽象基类"""

    def __init__(self, kl_pd_buy, factor_name):
        """
        :param kl_pd_buy: 交易当日的交易数据
        :param factor_name: FactorBuyBases子类实例对象的factor_name
        """
        self.buy_price = np.inf
        self.kl_pd_buy = kl_pd_buy
        self.factor_name = factor_name

    def fit(self):
        """做基础验证比如今天是否停盘后调用fit_price"""
        if self.kl_pd_buy.empty or self.kl_pd_buy.volume == 0:
            # 买入时正无穷为放弃单子
            return np.inf

        return self.fit_price()

    @abstractmethod
    def fit_price(self):
        """
        子类主要需要实现的函数，决策交易当日的最终买入价格
        :return: 最终决策的当前交易买入价格
        """
        pass


"""是否开启涨停板滑点买入价格特殊处理, 默认关闭，外部修改如：slippage.sbb.g_enable_limit_up = True"""
g_enable_limit_up = False
"""
    初始设定涨停板买入成交概率100%，这里也可以在计算完一次概率后，再使用成交量做二次概率计算，
    外部修改如：slippage.sbb.g_limit_up_deal_chance = 0.5，即修改为买入概率50%
"""
g_limit_up_deal_chance = 1
"""在集合竞价阶段价格已经达成涨停的情况下买入成功的概率，默认0.2, 即20%成功概率"""
g_pre_limit_up_rate = 0.2


def slippage_limit_up(func):
    """
        针对a股涨停板买入价格决策的装饰器，子类可选择装饰与不装饰在fit_price上
        如果是实盘策略中，使用分钟k线，及日内择时策略，即不需特别处理。
        回测中需要特别处理，处理买入成功概率，根据概率决定是否能买入，
        及涨停下的买入价格决策，涨停下买入价格模型为，越靠近涨停价格
        买入成交概率越大，即在涨停下预期以靠近涨停价格买入，缺点是使用了随机数，
        导致回测结果将出现不一致的情况
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if g_enable_limit_up and self.kl_pd_buy.p_change >= 10 and self.kl_pd_buy.high == self.kl_pd_buy.close:
            """
                涨停板命中后需要根据涨停板买入成交概率(g_limit_up_deal_chance)来作为
                二项式分布的概率值计算买入成功概率
            """
            if self.kl_pd_buy.high == self.kl_pd_buy.low:
                # 10个点，且最高＝最低，即a股在集合竞价阶段达成涨停，买入成功概率降低到g_limit_up_deal_chance * 0.2
                # TODO 这个概率最好使用成交量当日来计算出来
                limit_up_deal_chance = g_limit_up_deal_chance * g_pre_limit_up_rate
            else:
                limit_up_deal_chance = g_limit_up_deal_chance

            deal = np.random.binomial(1, limit_up_deal_chance)
            if deal:
                if self.kl_pd_buy.high == self.kl_pd_buy.low:
                    return self.kl_pd_buy.high

                # 买入成功后需要进一步决策价位，首选arange出一个从低到涨停价格的序列，间隔0.01
                price_lh = np.arange(self.kl_pd_buy.low, self.kl_pd_buy.high,
                                     0.01)
                # 构造概率序列，可以使用其它比如指数分布等，提高以涨停价格买入的概率，这里只使用最简单方式
                lh_chance = np.linspace(0, 1, len(price_lh))
                """
                    计算出对应的概率, 这里的概率分布并不陡峭，即涨停价格附近权重并不是很高，
                    可以使用如：np.power(price_hl, len(price_hl) / 2) / np.power(price_hl, len(price_hl) / 2).sum()
                    来进一步提升涨跌板附近价格的买入权重
                """
                # noinspection PyUnresolvedReferences
                p = lh_chance / lh_chance.sum()
                # 最后使用随机加权概率抽取，选中一个买入涨停价格
                return np.random.choice(price_lh, 1, p=p)[0]
            # 没能成交返回正无穷
            return np.inf
        else:
            return func(self, *args, **kwargs)

    return wrapper



# -*- encoding:utf-8 -*-
import pdb
import numpy as np
import pandas as pd
from abc import ABCMeta, abstractmethod
from collections import namedtuple
from ultron.factor.fitness.metrics import Metrics
from ultron.ump.core.fixes import six


class ScoreTuple(
        namedtuple('ScoreTuple', ('name', 'factors', 'returns', 'freq'))):
    __slots__ = ()

    def __repr__(self):
        return "name:{}\nfreq:{}\nfactors:{}\returns:{}".format(
            self.name, self.freq,
            self.factors.info() if self.factors is not None else 'zero order',
            self.returns.info() if self.returns is not None else 'zero order')


class BaseScorer(six.with_metaclass(ABCMeta, object)):

    def __init__(self, score_tuple_array, *arg, **kwargs):

        self.score_tuple_array = score_tuple_array
        self.score_dict = {}
        self.weights_cnt = -1

        self._init_self_begin(arg, *arg, **kwargs)

        if not hasattr(self, 'select_score_func'):
            raise RuntimeError('_init_self_begin must set select_score_func')
        if not hasattr(self, 'columns_name'):
            raise RuntimeError('_init_self_begin must set columns_name')

        # 如果有设置权重就分配权重否则等权重
        if 'weights' in kwargs and kwargs['weights'] is not None and len(
                kwargs['weights']) == self.weights_cnt:
            self.weights = kwargs['weights']
        else:
            self.weights = self.weights_cnt * [
                1. / self.weights_cnt,
            ]

        if 'metrics_class' in kwargs and kwargs['metrics_class'] is not None \
                and issubclass(kwargs['metrics_class'], Metrics):
            self.metrics_class = kwargs['metrics_class']
        else:
            self.metrics_class = Metrics

        for ind, score_tuple in enumerate(self.score_tuple_array):
            metrics = self.metrics_class(returns=score_tuple.returns,
                                         factors=score_tuple.factors)
            if metrics.valid:
                result = metrics.fit_metrics()
                self.score_dict["{0}_{1}_{2}".format(
                    score_tuple.name, result.long_evaluate.category,
                    result.long_evaluate.freq)] = self.select_score_func(
                        result.long_evaluate)

                self.score_dict["{0}_{1}_{2}".format(
                    score_tuple.name, result.short_evaluate.category,
                    result.short_evaluate.freq)] = self.select_score_func(
                        result.short_evaluate)

                self.score_dict["{0}_{1}_{2}".format(
                    score_tuple.name, result.both_evaluate.category,
                    result.both_evaluate.freq)] = self.select_score_func(
                        result.both_evaluate)

        # 将score_dict转换DataFrame并且转制
        self.score_pd = pd.DataFrame(self.score_dict).T
        # 设置度量指标名称
        self.score_pd.columns = self.columns_name

        self._init_self_end(arg, *arg, **kwargs)

        # 分数每一项都由0-1
        score_ls = np.linspace(0, 1, self.score_pd.shape[0])
        for cn in self.columns_name:
            # 每一项的结果rank后填入对应项
            score = score_ls[(self.score_pd[cn].fillna(0).rank().values -
                              1).astype(int)]
            self.score_pd['score_' + cn] = score

        scores = self.score_pd.filter(regex='score_*')
        # 根据权重计算最后的得分
        self.score_pd['score'] = scores.apply(lambda s:
                                              (s * self.weights).sum(),
                                              axis=1)

    @abstractmethod
    def _init_self_begin(self, *arg, **kwargs):
        """子类需要实现，设置度量项抽取函数select_score_func，度量名称columns_name，weights_cnt"""
        pass

    @abstractmethod
    def _init_self_end(self, *arg, **kwargs):
        """子类需要实现，一般的任务是将score_pd中需要反转的度量结果进行反转"""
        pass

    def fit_score(self):
        """对度量结果按照score排序，返回排序后的score列"""
        self.score_pd.sort_values(by='score', inplace=True)
        return self.score_pd['score']

    def __call__(self):
        """call self.fit_score"""
        return self.fit_score()


class DemoScorer(BaseScorer):

    def _init_self_begin(self, *arg, **kwargs):
        """胜率，策略收益，换手率组成select_score_func"""
        self.select_score_func = lambda metrics: [
            metrics.win_rate, metrics.returns_mean, metrics.turnover
        ]
        self.columns_name = ['win_rate', 'returns_mean', 'turnover']
        self.weights_cnt = len(self.columns_name)

    def _init_self_end(self, *arg, **kwargs):
        """
        _init_self_end这里一般的任务是将score_pd中需要反转的反转，默认是数据越大越好，有些是越小越好，
        类似make_scorer(xxx, greater_is_better=True)中的参数greater_is_better的作用：

                            sign = 1 if greater_is_better else -1
        """
        pass


# -*- coding: utf-8 -*-

import unittest

from ultron.optimize.model.linearmodel import ConstLinearModel

class ModelBase(unittest.TestCase):
    def test_simple_model_features(self):
        model = ConstLinearModel(features=['c', 'b', 'a'])
        self.assertListEqual(['a', 'b', 'c'], model.features)



# -*- coding: utf-8 -*-
import os, pdb, itertools, copy, datetime, sys, json
import pandas as pd

os.environ['ULTRON_DATA'] = 'keim'
from ultron.env import *
from ultron.sentry.api import *
from ultron.optimize.model.treemodel import RandomForestRegressor
from ultron.optimize.model.treemodel import RandomForestClassifier

enable_example_env()

train_data = pd.read_csv(os.path.join(g_project_data, 'train_datas.csv'),
                         index_col=0)

featrues = [
    MSUM(10, 'BM_MainFar_80D'), 'BM_RecentFar_40D',
    'WeightShortVolRelTotIntChg'
]
train_data['trade_date'] = pd.to_datetime(train_data['trade_date'])
model = RandomForestClassifier(features=featrues)
#model.formulas.transform('code', cmt)
#pdb.set_trace()
#model.fit(train_data[featrues].fillna(0),
#          train_data['signal'].fillna(0).values)
#print('--->')


import os, time
import datetime as dt
import sqlalchemy
import numpy as np
import pandas as pd
import statsmodels.api as sm


class OnlySideWeighted(object):

    def __init__(self):
        self._industry_styles = [
            'Bank', 'RealEstate', 'Health', 'Transportation', 'Mining',
            'NonFerMetal', 'HouseApp', 'LeiService', 'MachiEquip', 'BuildDeco',
            'CommeTrade', 'CONMAT', 'Auto', 'Textile', 'FoodBever',
            'Electronics', 'Computer', 'LightIndus', 'Utilities', 'Telecom',
            'AgriForest', 'CHEM', 'Media', 'IronSteel', 'NonBankFinan',
            'ELECEQP', 'AERODEF', 'Conglomerates'
        ]
        self._risk_styles = [
            'BETA', 'MOMENTUM', 'SIZE', 'EARNYILD', 'RESVOL', 'GROWTH', 'BTOP',
            'LEVERAGE', 'LIQUIDTY', 'SIZENL'
        ]

    #暂时写此处，去极值
    def winsorize(self,
                  total_data,
                  method='sigma',
                  limits=(3.0, 3.0),
                  drop=True):
        se = total_data.copy()
        if method == 'quantile':
            down, up = se.quantile([limits[0], 1.0 - limits[1]])
        elif method == 'sigma':
            std, mean = se.std(), se.mean()
            down, up = mean - limits[0] * std, mean + limits[1] * std

        if drop:
            se[se < down] = np.NaN
            se[se > up] = np.NaN
        else:
            se[se < down] = down
            se[se > up] = up
        return se

    #标准化
    def standardize(self, total_data):
        try:
            res = (total_data - np.nanmean(total_data)) / np.nanstd(total_data)
        except:
            res = pd.Series(data=np.NaN, index=total_data.index)
        return res

    #中性化
    def neutralize(self, total_data, risk_df):
        se = total_data.dropna()
        # se = total_data.copy()
        risk = risk_df.loc[se.index, :]
        # use numpy for neu, which is faster
        x = np.linalg.lstsq(risk.values, np.matrix(se).T)[0]
        se_neu = se - risk.dot(x)[0]

        return se_neu

    def top_equal_weights(self, group, top='top', top_pct=0.2):
        if top == 'bottom':
            group = -1.0 * group
        group = group.rank(pct=True)
        group[group < 1.0 - top_pct] = 0.0
        group[group >= 1.0 -
              top_pct] = 1.0 / len(group[group >= 1.0 - top_pct])
        return group

    def calc_stats(self, returns_df):
        # 总体指标
        returns_se, turnover_se = returns_df['returns'], returns_df['turnover']

        ir = returns_se.mean() / returns_se.std()
        sharpe = np.sqrt(252) * ir
        turnover = turnover_se.mean()
        returns = returns_se.sum() * 252 / len(returns_se)
        fitness = sharpe * np.sqrt(abs(returns) / turnover)
        margin = returns_se.sum() / turnover_se.sum()
        stats_se = pd.Series({
            'status': 1,
            'ir': ir,
            'sharpe': sharpe,
            'turnover': turnover,
            'returns': returns,
            'fitness': fitness,
            'margin': margin
        })
        return stats_se

    def returns(self, factor_se, forward_returns, init_capital=100000):
        # 超额收益，此处benchmark采用全股票池算数均值
        forward_excess_returns = forward_returns.groupby(
            level=['trade_date']).apply(lambda x: x - x.mean())
        res_dict = {}
        for top in ['top', 'bottom']:
            weights = factor_se.groupby(level=['trade_date']).apply(
                lambda x: self.top_equal_weights(x, top=top, top_pct=0.2))
            weighted_returns = forward_excess_returns.multiply(weights, axis=0)
            factor_ret_se = weighted_returns.groupby(level='trade_date').sum()
            turnover_se = weights.unstack().diff().abs().sum(axis=1)
            res_dict[top] = pd.DataFrame({
                'returns': factor_ret_se,
                'turnover': turnover_se
            })
        return res_dict

    def run(self,
            factor_data,
            risk_data,
            forward_returns,
            default_value,
            factor_name,
            horizon=1,
            method='quantile',
            up_limit=0.025,
            down_limit=0.025,
            init_capital=100000):
        """
        参数：
            horizon: 调仓期，按照交易日计算。
        """
        factor_se = factor_data.set_index(['trade_date', 'code'])
        risk_se = risk_data.set_index(['trade_date', 'code'])
        forward_returns = forward_returns.set_index(['trade_date', 'code'])
        risk_df = risk_se.reindex(factor_se.index)[self._industry_styles +
                                                   ['SIZE']]
        risk_df.dropna(inplace=True)
        factor_se = factor_se.loc[risk_df.index][factor_name]
        returns = forward_returns.loc[risk_df.index]

        factor_se = factor_se.groupby(
            level='trade_date').apply(lambda x: self.winsorize(
                x, method=method, limits=(up_limit, down_limit)))
        grouped = factor_se.groupby(level='trade_date')
        res = []
        for trade_date, g in grouped:
            neu = self.neutralize(g, risk_df)
            res.append(neu)
        factor_se = pd.concat(res, axis=0)

        factor_se = factor_se.groupby(
            level='trade_date').apply(lambda x: self.standardize(x))

        #判读覆盖率低于某个阈值直接设置无效
        cover_rate = 1 - len(factor_se[np.isnan(
            factor_se.values)]) / len(factor_se)
        if cover_rate < 0.75:
            return {'status': 0, 'fitness': default_value}
        # top 20% 等权方法计算的因子收益数据
        res_dict = self.returns(factor_se, forward_returns['ret'])

        #计算指标
        stats_df = pd.DataFrame(
            {x: self.calc_stats(res_dict[x])
             for x in res_dict.keys()})
        if stats_df.top.returns > stats_df.bottom.returns and stats_df.top.returns > 0:
            return stats_df.top.to_dict()
        elif stats_df.bottom.returns > stats_df.top.returns and stats_df.bottom.returns > 0:
            return stats_df.bottom.to_dict()
        else:
            return {'status': 0, 'fitness': default_value}



# -*- encoding:utf-8 -*-
from ultron.factor.fitness.metrics import Metrics


def metrics_fitness(factor_data, total_data, factor_sets, custom_params,
                    default_value):
    returns = total_data[['trade_date', 'code', 'nxt1_ret']]
    factor_data = factor_data.reset_index()
    data = factor_data.merge(returns, on=['trade_date', 'code'])
    data = data.set_index(['trade_date', 'code']).dropna(subset=['nxt1_ret'])
    factors_socre = data['transformed'].unstack()
    yield_score = data['nxt1_ret'].unstack()
    ms = Metrics(returns=yield_score, factors=factors_socre)
    evaluate_name = custom_params['evaluate']
    method_name = custom_params['method']
    result = ms.fit_metrics()
    fitness = float(
        result.__getattribute__(f"{evaluate_name}").__getattribute__(
            f"{method_name}"))
    return fitness, result



# -*- coding: utf-8 -*-
import numpy as np
from sklearn.linear_model import LinearRegression as LinearRegressionImpl
from sklearn.linear_model import Lasso as LassoRegressionImpl
from sklearn.linear_model import LogisticRegression as LogisticRegressionImpl
from sklearn.linear_model import BayesianRidge as BayesianRidgeImpl
from sklearn.linear_model import Ridge as RidgeImpl
from sklearn.linear_model import TweedieRegressor as TweedieRegressorImpl
from sklearn.linear_model import HuberRegressor as HuberRegressorImpl
from sklearn.linear_model import SGDRegressor as SGDRegressorImpl
from sklearn.linear_model import PassiveAggressiveRegressor as PassiveAggressiveRegressorImpl
from sklearn.linear_model import TheilSenRegressor as TheilSenRegressorImpl
from ultron.optimize.model.modelbase import create_model_base


class ConstLinearModelImpl(object):

    def __init__(self, weights: np.ndarray = None):
        self.weights = weights.flatten()

    def fit(self, x: np.ndarray, y: np.ndarray):
        raise NotImplementedError(
            "Const linear model doesn't offer fit methodology")

    def predict(self, x: np.ndarray):
        return x @ self.weights

    def score(self, x: np.ndarray, y: np.ndarray) -> float:
        y_hat = self.predict(x)
        y_bar = y.mean()
        ssto = ((y - y_bar)**2).sum()
        sse = ((y - y_hat)**2).sum()
        return 1. - sse / ssto


class ConstLinearModel(create_model_base()):

    def __init__(self, features=None, weights: dict = None, fit_target=None):
        super().__init__(features=features, fit_target=fit_target)
        if features is not None and weights is not None:
            assert (len(features) == len(weights), ValueError,
                    "length of features is not equal to length of weights")
        if weights:
            self.impl = ConstLinearModelImpl(
                np.array([weights[name] for name in self.features]))

    def save(self):
        model_desc = super().save()
        model_desc['weight'] = list(self.impl.weights)
        return model_desc

    @classmethod
    def load(cls, model_desc: dict):
        return super().load(model_desc)

    @property
    def weights(self):
        return self.impl.weights.tolist()


class LinearRegression(create_model_base('sklearn')):

    def __init__(self,
                 features=None,
                 fit_intercept: bool = False,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = LinearRegressionImpl(fit_intercept=fit_intercept, **kwargs)

    def save(self) -> dict:
        model_desc = super().save()
        model_desc['weight'] = self.impl.coef_.tolist()
        return model_desc

    @property
    def weights(self):
        return self.impl.coef_.tolist()


class LassoRegression(create_model_base('sklearn')):

    def __init__(self,
                 alpha=0.01,
                 features=None,
                 fit_intercept: bool = False,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = LassoRegressionImpl(alpha=alpha,
                                        fit_intercept=fit_intercept,
                                        **kwargs)

    def save(self) -> dict:
        model_desc = super().save()
        model_desc['weight'] = self.impl.coef_.tolist()
        return model_desc

    @property
    def weights(self):
        return self.impl.coef_.tolist()


class BayesianRegression(create_model_base('sklearn')):

    def __init__(self,
                 alpha=0.01,
                 features=None,
                 fit_intercept: bool = False,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = BayesianRidgeImpl(alpha_1=alpha,
                                      fit_intercept=fit_intercept,
                                      **kwargs)

    def save(self) -> dict:
        model_desc = super().save()
        model_desc['weight'] = self.impl.coef_.tolist()
        return model_desc

    @property
    def weights(self):
        return self.impl.coef_.tolist()


class RidgeRegression(create_model_base('sklearn')):

    def __init__(self,
                 alpha=0.01,
                 features=None,
                 fit_intercept: bool = False,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = RidgeImpl(alpha=alpha,
                              fit_intercept=fit_intercept,
                              **kwargs)

    def save(self) -> dict:
        model_desc = super().save()
        model_desc['weight'] = self.impl.coef_.tolist()
        return model_desc

    @property
    def weights(self):
        return self.impl.coef_.tolist()


class TweedieRegression(create_model_base('sklearn')):

    def __init__(self,
                 alpha=0.01,
                 features=None,
                 fit_intercept: bool = False,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = TweedieRegressorImpl(alpha=alpha,
                                         fit_intercept=fit_intercept,
                                         **kwargs)

    def save(self) -> dict:
        model_desc = super().save()
        model_desc['weight'] = self.impl.coef_.tolist()
        return model_desc

    @property
    def weights(self):
        return self.impl.coef_.tolist()


class HuberRegression(create_model_base('sklearn')):

    def __init__(self,
                 alpha=0.01,
                 features=None,
                 fit_intercept: bool = False,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = HuberRegressorImpl(alpha=alpha,
                                       fit_intercept=fit_intercept,
                                       **kwargs)

    def save(self) -> dict:
        model_desc = super().save()
        model_desc['weight'] = self.impl.coef_.tolist()
        return model_desc

    @property
    def weights(self):
        return self.impl.coef_.tolist()


class SGDRegression(create_model_base('sklearn')):

    def __init__(self,
                 alpha=0.001,
                 features=None,
                 fit_intercept: bool = False,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = SGDRegressorImpl(alpha=alpha,
                                     fit_intercept=fit_intercept,
                                     **kwargs)

    def save(self) -> dict:
        model_desc = super().save()
        model_desc['weight'] = self.impl.coef_.tolist()
        return model_desc

    @property
    def weights(self):
        return self.impl.coef_.tolist()


class PassiveAggressiveRegression(create_model_base('sklearn')):

    def __init__(self,
                 features=None,
                 fit_intercept: bool = False,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = PassiveAggressiveRegressorImpl(fit_intercept=fit_intercept,
                                                   **kwargs)

    def save(self) -> dict:
        model_desc = super().save()
        model_desc['weight'] = self.impl.coef_.tolist()
        return model_desc

    @property
    def weights(self):
        return self.impl.coef_.tolist()


class TheilSenRegression(create_model_base('sklearn')):

    def __init__(self,
                 features=None,
                 fit_intercept: bool = False,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = TheilSenRegressorImpl(fit_intercept=fit_intercept,
                                          **kwargs)

    def save(self) -> dict:
        model_desc = super().save()
        model_desc['weight'] = self.impl.coef_.tolist()
        return model_desc

    @property
    def weights(self):
        return self.impl.coef_.tolist()


class LogisticRegression(create_model_base('sklearn')):

    def __init__(self,
                 features=None,
                 fit_intercept: bool = False,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = LogisticRegressionImpl(fit_intercept=fit_intercept,
                                           **kwargs)

    def save(self) -> dict:
        model_desc = super().save()
        model_desc['weight'] = self.impl.coef_.tolist()
        return model_desc

    @property
    def weights(self):
        return self.impl.coef_.tolist()
